"""
This module offers different interpreters for RDST classifiers and transformers. 
"""

from .rdst_interpreter import RDST_Ridge_interpreter, RDST_interpreter

__author__ = 'Antoine Guillaume antoine.guillaume45@gmail.com'

__all__ = ["RDST_Ridge_interpreter","RDST_interpreter"]


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Dec  1 09:22:29 2022

@author: lifo
"""
import pandas as pd

import numpy as np

import seaborn as sns

import operator

import math

import networkx

from scipy.stats import wilcoxon
from scipy.stats import friedmanchisquare

from matplotlib import pyplot as plt

# TODO: refactor methods

def pairwise_plot(
    df, baseline, margin=0.015, y_min=0, y_max=1, show_names_above=0.7,
    max_ncols=2, sns_context='talk', figsize=None, dpi=None, show_win_areas=False
):
    """
    Make pairwise plots using a dataframe with columns as a model performance
    and index as the dataset name.

    Parameters
    ----------
    df : DataFrame
        A dataframe containing the performance of models as columns, and the
        dataset names as index.
    baseline : str
        Name of the columns to use as baseline on the y-axis.
    margin : float, optional
        Margin separating the x=y line. If the difference between two models is 
        within +/- margin, this dataset will be considered as a draw.
        The default is 0.025.
    y_min : float, optional
        Minimum value on the axes. The default is 0.
    y_max : TYPE, optional
        Maximum value on the axes. The default is 1.
    show_names_above : float, optional
        If a difference between two models on a data is above this value,
        corresponding index name will be shown above the point.
        The default is 0.7.
    max_ncols : int, optional
        The maximum number of columns to draw. The number of rows is equal to 
        the number of columns in df - 1 divided by max_ncols, rounded up.
        The default is 2.
    sns_context : str, optional
        Type of seaborn context to apply, use None to not use sns style.
        The default is 'talk'.
    figsize : set, optional
        A two dimensional set as (x,y), with x the width of the figure and
        y its height. The default is None and will automatically adjust the 
        size of the figure.

    Returns
    -------
    fig : matplotlib.Figure
        The resulting pairwise plot.

    """
    if sns_context is not None:
        sns.set()
        sns.set_context(sns_context)
    competitors = df.columns.difference([baseline])
    if len(competitors)<max_ncols:
        ncols=len(competitors)
        nrows=1
    else:
        ncols = max_ncols
        nrows = int(np.ceil(len(competitors)/ncols))
    if figsize is None:
        figsize = (7.5*ncols, 7.5*nrows)
    fig, ax = plt.subplots(
        ncols=ncols, nrows=nrows, figsize=figsize, sharey=True, dpi=dpi
    )
    props = dict(boxstyle='round', facecolor='wheat', alpha=0.75)
    for i, comp in enumerate(competitors):
        if nrows > 1:
            limp = (y_max-y_min)*0.05
            ax[i // ncols, i % ncols].set_xlim(y_min-limp, y_max+limp)
            ax[i // ncols, i % ncols].set_ylim(y_min-limp, y_max+limp)
            ax[i // ncols, i %
                ncols].plot([y_min, y_max], [y_min, y_max], color='r')
            ax[i // ncols, i % ncols].plot([y_min, y_max-margin], [
                                           y_min + margin, y_max], color='orange', linestyle='--', alpha=0.75)
            ax[i // ncols, i % ncols].plot([y_min + margin, y_max], [
                                           y_min, y_max-margin], color='orange', linestyle='--', alpha=0.75)
            ax[i // ncols, i % ncols].annotate(
                '{}'.format(margin),
                (y_min + margin/2 + 0.0125, y_min+margin/2 - 0.01),
                fontsize=12
            )
            x = df[comp].values
            y = df[baseline].values
            ax[i // ncols, i % ncols].scatter(x, y, s=15, alpha=0.75, zorder=2)
            if np.any(np.abs(x-y)>=show_names_above):
                for j in range(df.shape[0]):
                    x1 = df.iloc[j][comp]
                    y1 = df.iloc[j][baseline]
                    if abs(x1 - y1) > show_names_above:
                        ax[i // ncols, i % ncols].annotate(
                            "{}".format(df.index[j]),
                            (x1, y1),
                            fontsize=14,
                            bbox=dict(boxstyle="round", alpha=0.1),
                        )
            if i % ncols == 0:
                ax[i // ncols, i % ncols].set_ylabel(baseline)
            ax[i // ncols, i % ncols].set_xlabel(comp)
    
            textstr = 'W - D - L\n{} - {} - {}'.format(sum(x+margin < y), sum(
                (x <= y+margin) & (y-margin <= x)), sum(x-margin > y))
            ax[i // ncols, i % ncols].text(0.05, 0.95, textstr, transform=ax[i // ncols, i % ncols].transAxes, fontsize=14,
                                           verticalalignment='top', bbox=props)
        if nrows == 1:
            if ncols==1:
            
                limp = (y_max-y_min)*0.05
                ax.set_xlim(y_min-limp, y_max+limp)
                ax.set_ylim(y_min-limp, y_max+limp)
                ax.plot([y_min, y_max], [y_min, y_max], color='r')
                ax.plot([y_min, y_max-margin], [
                                               y_min + margin, y_max], color='orange', linestyle='--', alpha=0.75)
                ax.plot([y_min + margin, y_max], [
                                               y_min, y_max-margin], color='orange', linestyle='--', alpha=0.75)
                
                ax.annotate(
                    '{}'.format(margin),
                    (y_min + margin/2 + 0.0125, y_min+margin/2 - 0.01),
                    fontsize=12
                )
                x = df[comp].values
                y = df[baseline].values
                ax.scatter(x, y, s=15, alpha=0.75, zorder=2)
                if np.any(np.abs(x-y)>=show_names_above):
                    for j in range(df.shape[0]):
                        x1 = df.iloc[j][comp]
                        y1 = df.iloc[j][baseline]
                        if abs(x1 - y1) > show_names_above:
                            ax.annotate(
                                "{}".format(df.index[j]),
                                (x1, y1),
                                fontsize=14,
                                bbox=dict(boxstyle="round", alpha=0.1)
                            )
                if i % ncols == 0:
                    ax.set_ylabel(baseline)
                ax.set_xlabel(comp)
    
                textstr = 'W - D - L\n{} - {} - {}'.format(sum(x+margin < y), sum(
                    (x <= y+margin) & (y-margin <= x)), sum(x-margin > y))
                ax.text(0.05, 0.95, textstr, transform=ax.transAxes, fontsize=14,
                                               verticalalignment='top', bbox=props)
            else:
                limp = (y_max-y_min)*0.05
                ax[i % ncols].set_xlim(y_min-limp, y_max+limp)
                ax[i % ncols].set_ylim(y_min-limp, y_max+limp)
                ax[i %
                    ncols].plot([y_min, y_max], [y_min, y_max], color='r')
                ax[i % ncols].plot([y_min, y_max-margin], [
                                               y_min + margin, y_max], color='orange', linestyle='--', alpha=0.75)
                ax[i % ncols].plot([y_min + margin, y_max], [
                                               y_min, y_max-margin], color='orange', linestyle='--', alpha=0.75)
                
                ax[i % ncols].annotate(
                    '{}'.format(margin),
                    (y_min + margin/2 + 0.0125, y_min+margin/2 - 0.01),
                    fontsize=12
                )
                x = df[comp].values
                y = df[baseline].values
                ax[i % ncols].scatter(x, y, s=15, alpha=0.75, zorder=2)
                if np.any(np.abs(x-y)>=show_names_above):
                    for j in range(df.shape[0]):
                        x1 = df.iloc[j][comp]
                        y1 = df.iloc[j][baseline]
                        if abs(x1 - y1) > show_names_above:
                            ax[i % ncols].annotate(
                                "{}".format(df.index[j]),
                                (x1, y1),
                                fontsize=14,
                                bbox=dict(boxstyle="round", alpha=0.1)
                            )
                if i % ncols == 0:
                    ax[i % ncols].set_ylabel(baseline)
                ax[i % ncols].set_xlabel(comp)
        
                textstr = 'W - D - L\n{} - {} - {}'.format(sum(x+margin < y), sum(
                    (x <= y+margin) & (y-margin <= x)), sum(x-margin > y))
                ax[i % ncols].text(0.05, 0.95, textstr, transform=ax[i % ncols].transAxes, fontsize=14,
                                               verticalalignment='top', bbox=props)
    return fig

def draw_cd_diagram(df_perf=None, alpha=0.05, title=None, width=10, labels=False, path=None, highlight=None):
    """
    Draws the critical difference diagram given the list of pairwise classifiers that are
    significant or not
    """
    p_values, average_ranks, _ = _wilcoxon_holm(df_perf=df_perf, alpha=alpha)
    if p_values is not None:
        _graph_ranks(average_ranks.values, average_ranks.keys(), p_values, 
                    cd=None, reverse=True, width=width, textspace=1.25, labels=labels,
                    highlight=highlight)

        font = {'family': 'sans-serif',
                'color': 'black',
                'weight': 'normal',
                'size': 18,
                }
        if title:
            plt.title(title, fontdict=font, y=0.9, x=0.5)
        if path is None:
            plt.savefig('cd-diagram.png', bbox_inches='tight')

def _form_cliques(p_values, nnames):
    """
    This method forms the cliques
    """
    # first form the numpy matrix data
    m = len(nnames)
    g_data = np.zeros((m, m), dtype=np.int64)
    for p in p_values:
        if p[3] == False:
            i = np.where(nnames == p[0])[0][0]
            j = np.where(nnames == p[1])[0][0]
            min_i = min(i, j)
            max_j = max(i, j)
            g_data[min_i, max_j] = 1

    g = networkx.Graph(g_data)
    return networkx.find_cliques(g)

def _wilcoxon_holm(alpha=0.05, df_perf=None):
    """
    Applies the wilcoxon signed rank test between each pair of algorithm and then use Holm
    to reject the null's hypothesis
    """
    # count the number of tested datasets per classifier
    df_counts = pd.DataFrame({'count': df_perf.groupby(
        ['classifier_name']).size()}).reset_index()
    # get the maximum number of tested datasets
    max_nb_datasets = df_counts['count'].max()
    # get the list of classifiers who have been tested on nb_max_datasets
    classifiers = list(df_counts.loc[df_counts['count'] == max_nb_datasets]
                       ['classifier_name'])
    # test the null hypothesis using friedman before doing a post-hoc analysis
    friedman_p_value = friedmanchisquare(*(
        np.array(df_perf.loc[df_perf['classifier_name'] == c]['accuracy'])
        for c in classifiers))[1]
    print(friedman_p_value)

    # get the number of classifiers
    m = len(classifiers)
    # init array that contains the p-values calculated by the Wilcoxon signed rank test
    p_values = []
    # loop through the algorithms to compare pairwise
    for i in range(m - 1):
        # get the name of classifier one
        classifier_1 = classifiers[i]
        # get the performance of classifier one
        perf_1 = np.array(
            df_perf.loc[df_perf['classifier_name'] == classifier_1]['accuracy'], dtype=np.float64)
        for j in range(i + 1, m):
            # get the name of the second classifier
            classifier_2 = classifiers[j]
            # get the performance of classifier one
            perf_2 = np.array(df_perf.loc[df_perf['classifier_name'] == classifier_2]
                              ['accuracy'], dtype=np.float64)
            # calculate the p_value
            p_value = wilcoxon(perf_1, perf_2, zero_method='pratt')[1]
            # appen to the list
            p_values.append((classifier_1, classifier_2, p_value, False))
    # get the number of hypothesis
    k = len(p_values)
    # sort the list in acsending manner of p-value
    p_values.sort(key=operator.itemgetter(2))

    # loop through the hypothesis
    for i in range(k):
        # correct alpha with holm
        new_alpha = float(alpha / (k - i))
        # test if significant after holm's correction of alpha
        if p_values[i][2] <= new_alpha:
            p_values[i] = (p_values[i][0], p_values[i]
                           [1], p_values[i][2], True)
        else:
            # stop
            break
    # compute the average ranks to be returned (useful for drawing the cd diagram)
    # sort the dataframe of performances
    sorted_df_perf = df_perf.loc[df_perf['classifier_name'].isin(classifiers)]. \
        sort_values(['classifier_name', 'dataset_name'])
    # get the rank data
    rank_data = np.array(sorted_df_perf['accuracy']).reshape(
        m, max_nb_datasets)

    # create the data frame containg the accuracies
    df_ranks = pd.DataFrame(data=rank_data, index=np.sort(
        classifiers), columns=np.unique(sorted_df_perf['dataset_name']))

    # average the ranks
    average_ranks = df_ranks.rank(ascending=False).mean(
        axis=1).sort_values(ascending=False)
    # return the p-values and the average ranks
    return p_values, average_ranks, max_nb_datasets



def _graph_ranks(avranks, names, p_values, cd=None, cdmethod=None, lowv=None, highv=None, highlight=None,
                width=6, textspace=1, reverse=False, filename=None, labels=False, **kwargs):
    """
    Draws a CD graph, which is used to display  the differences in methods'
    performance. See Janez Demsar, Statistical Comparisons of Classifiers over
    Multiple Data Sets, 7(Jan):1--30, 2006.
    Needs matplotlib to work.
    The image is ploted on `plt` imported using
    `import matplotlib.pyplot as plt`.
    Args:
        avranks (list of float): average ranks of methods.
        names (list of str): names of methods.
        cd (float): Critical difference used for statistically significance of
            difference between methods.
        cdmethod (int, optional): the method that is compared with other methods
            If omitted, show pairwise comparison of methods
        lowv (int, optional): the lowest shown rank
        highv (int, optional): the highest shown rank
        width (int, optional): default width in inches (default: 6)
        textspace (int, optional): space on figure sides (in inches) for the
            method names (default: 1)
        reverse (bool, optional): if set to `True`, the lowest rank is on the
            right (default: `False`)
        filename (str, optional): output file name (with extension). If not
            given, the function does not write a file.
        labels (bool, optional): if set to `True`, the calculated avg rank
        values will be displayed
    """
    width = float(width)
    textspace = float(textspace)
    
    def lloc(_list, n):
         """
         List location in list of list structure.
         Enable the use of negative locations:
         -1 is the last element, -2 second last...
         """
         if n < 0:
             return len(_list[0]) + n
         return n
     
    def nth(_list, n):
        """
        Returns only nth elemnt in a list.
        """
        n = lloc(_list, n)
        return [a[n] for a in _list]

    def mxrange(lr):
        """
        Multiple xranges. Can be used to traverse matrices.
        This function is very slow due to unknown number of
        parameters.
        >>> mxrange([3,5])
        [(0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2)]
        >>> mxrange([[3,5,1],[9,0,-3]])
        [(3, 9), (3, 6), (3, 3), (4, 9), (4, 6), (4, 3)]
        """
        if len(lr):
            yield ()
        else:
            # it can work with single numbers
            index = lr[0]
            if isinstance(index, int):
                index = [index]
            for a in range(*index):
                for b in mxrange(lr[1:]):
                    yield tuple([a] + list(b))

    sums = avranks

    nnames = names
    ssums = sums

    if lowv is None:
        lowv = min(1, int(math.floor(min(ssums))))
    if highv is None:
        highv = max(len(avranks), int(math.ceil(max(ssums))))

    cline = 0.4

    k = len(sums)

    linesblank = 0
    scalewidth = width - 2 * textspace

    def rankpos(rank):
        if not reverse:
            a = rank - lowv
        else:
            a = highv - rank
        return textspace + scalewidth / (highv - lowv) * a

    distanceh = 0.25

    cline += distanceh

    # calculate height needed height of an image
    minnotsignificant = max(2 * 0.2, linesblank)
    height = cline + ((k + 1) / 2) * 0.2 + minnotsignificant

    fig = plt.figure(figsize=(width, height*1.05))
    fig.set_facecolor('white')
    ax = fig.add_axes([0, 0, 1, 1])  # reverse y axis
    ax.set_axis_off()

    hf = 1. / height  # height factor
    wf = 1. / width

    def hfl(_list):
        return [a * hf for a in _list]

    def wfl(_list):
        return [a * wf for a in _list]

    # Upper left corner is (0,0).
    ax.plot([0, 1], [0, 1], c="w")
    ax.set_xlim(0, 1)
    ax.set_ylim(1, 0)

    def line(l, color='k', **kwargs):
        """
        Input is a list of pairs of points.
        """
        ax.plot(wfl(nth(l, 0)), hfl(nth(l, 1)), color=color, **kwargs)

    def text(x, y, s, *args, **kwargs):
        ax.text(wf * x, hf * y, s, *args, **kwargs)

    line([(textspace, cline), (width - textspace, cline)], linewidth=2)

    bigtick = 0.3
    smalltick = 0.15
    linewidth = 2.0
    linewidth_sign = 4.0

    tick = None
    for a in list(np.arange(lowv, highv, 0.5)) + [highv]:
        tick = smalltick
        if a == int(a):
            tick = bigtick
        line([(rankpos(a), cline - tick / 2),
              (rankpos(a), cline)],
             linewidth=2)

    for a in range(lowv, highv + 1):
        text(rankpos(a), cline - tick / 2 - 0.05, str(a),
             ha="center", va="bottom", size=16)

    k = len(ssums)

    def filter_names(name):
        return name

    space_between_names = 0.24

    for i in range(math.ceil(k / 2)):
        chei = cline + minnotsignificant + i * space_between_names
        if nnames[i] == highlight:
            line([(rankpos(ssums[i]), cline),
                  (rankpos(ssums[i]), chei),
                  (textspace - 0.1, chei)],
                 linewidth=linewidth, color='red')
        else:
            line([(rankpos(ssums[i]), cline),
                  (rankpos(ssums[i]), chei),
                  (textspace - 0.1, chei)],
                 linewidth=linewidth)
        if labels:
            text(textspace + 0.3, chei - 0.075,
                 format(ssums[i], '.4f'), ha="right", va="center", size=10)
        if nnames[i] == highlight:
            text(textspace - 0.2, chei,
                 filter_names(nnames[i]), ha="right", va="center", size=18, color='red')
        else:
            text(textspace - 0.2, chei,
                 filter_names(nnames[i]), ha="right", va="center", size=18)

    for i in range(math.ceil(k / 2), k):
        chei = cline + minnotsignificant + (k - i - 1) * space_between_names
        if nnames[i] == highlight:
            line([(rankpos(ssums[i]), cline),
                  (rankpos(ssums[i]), chei),
                  (textspace + scalewidth + 0.1, chei)],
                 linewidth=linewidth, color='red')
        else:
            line([(rankpos(ssums[i]), cline),
                  (rankpos(ssums[i]), chei),
                  (textspace + scalewidth + 0.1, chei)],
                 linewidth=linewidth)
        if labels:
            text(textspace + scalewidth - 0.3, chei - 0.075,
                 format(ssums[i], '.4f'), ha="left", va="center", size=10)
        if nnames[i] == highlight:
            text(textspace + scalewidth + 0.2, chei, filter_names(nnames[i]),
                 ha="left", va="center", size=18, color='red')
        else:
            text(textspace + scalewidth + 0.2, chei, filter_names(nnames[i]),
                 ha="left", va="center", size=18)
    start = cline + 0.2
    side = -0.02
    height = 0.1

    # draw no significant lines
    # get the cliques
    cliques = _form_cliques(p_values, nnames)
    achieved_half = False
    print(nnames)
    for clq in cliques:
        if len(clq) == 1:
            continue
        print(clq)
        min_idx = np.array(clq).min()
        max_idx = np.array(clq).max()
        if min_idx >= len(nnames) / 2 and achieved_half == False:
            start = cline + 0.25
            achieved_half = True
        line([(rankpos(ssums[min_idx]) - side, start),
              (rankpos(ssums[max_idx]) + side, start)],
             linewidth=linewidth_sign)
        start += height


# -*- coding: utf-8 -*-
"""
@author: Antoine Guillaume
"""
from numpy.random import choice, uniform, random, seed
from numpy import (
    unique, where, percentile, int64, bool_, float64, concatenate, any as _any,
    dot, log2, floor_divide, zeros, floor, power, ones, cumsum, mean, std,
    arange
)

from convst.transformers._commons import (
    get_subsequence, compute_shapelet_dist_vector,
    apply_one_shapelet_one_sample_multivariate, _combinations_1d,
    generate_strides_2D, prime_up_to, choice_log
)

from numba import njit, prange

from convst import (
    __USE_NUMBA_CACHE__, __USE_NUMBA_FASTMATH__,
    __USE_NUMBA_NOGIL__, __USE_NUMBA_PARALLEL__
)


@njit(cache=__USE_NUMBA_CACHE__, nogil=__USE_NUMBA_NOGIL__)
def M_VL_init_random_shapelet_params(
    n_shapelets, shapelet_sizes, n_timestamps, p_norm, max_channels, prime_scheme
):
    """
    Initialize the parameters of the shapelets.    

    Parameters
    ----------
    n_shapelets : int
        Number of shapelet to initialize
    shapelet_sizes : array, shape=()
        Set of possible length for the shapelets
    n_timestamps : int
        Number of timestamps in the input data
    p_norm : float
        A value in the range [0,1] indicating the chance for each
        shapelet to use z-normalized distance
    max_channels : int
        The maximum number of features considered for one shapelet
    prime_scheme : bool
        Wheter to only consider prime numbers as possible dilation 

        
    Returns
    -------
    values : array, shape=(n_shapelet, max(shapelet_sizes))
        An initialized (empty) value array for each shapelet
    lengths : array, shape=(n_shapelet)
        The randomly initialized length of each shapelet
    dilations : array, shape=(n_shapelet)
        The randomly initialized dilation of each shapelet
    threshold : array, shape=(n_shapelet)
        An initialized (empty) value array for each shapelet
    normalize : array, shape=(n_shapelet)
        The randomly initialized normalization indicator of each shapelet
    channels : array, shape=(n_shapelet, n_features)
        The features considered by each shapelet
    """
    # Lengths of the shapelets
    lengths = choice(shapelet_sizes, size=n_shapelets).astype(int64)

    # Dilations
    upper_bounds = log2(floor_divide(n_timestamps - 1, lengths - 1))
    if prime_scheme:
        primes = prime_up_to(int64(2**upper_bounds.max()))
        dilations = zeros(n_shapelets, dtype=int64)
        #TODO : optimize to avoid recomputing choice log for all upper bounds
        #Loop on each unique upper bounds ?
        for i in prange(n_shapelets):
            shp_primes = primes[primes<=int64(2**upper_bounds[i])]
            dilations[i] = shp_primes[choice_log(shp_primes.shape[0], 1)[0]]
    else:
        powers = zeros(n_shapelets)
        for i in prange(n_shapelets):
            powers[i] = uniform(0, upper_bounds[i])
        dilations = floor(power(2, powers)).astype(int64)

    # Init threshold array
    threshold = zeros(n_shapelets)

    # channels (i.e. features)
    n_channels = choice(max_channels, size=n_shapelets)+1

    channel_ids = zeros(n_channels.sum(), dtype=int64)

    # Init values array
    values = zeros(
        int64(
            dot(lengths.astype(float64), n_channels.astype(float64))
        )
    )

    # Is shapelet using z-normalization ?
    normalize = random(size=n_shapelets)
    normalize = (normalize < p_norm)

    return values, lengths, dilations, threshold, normalize, n_channels, channel_ids

@njit(cache=__USE_NUMBA_CACHE__, parallel=__USE_NUMBA_PARALLEL__, nogil=__USE_NUMBA_NOGIL__)
def M_VL_generate_shapelet(
    X, y, n_shapelets, shapelet_sizes, r_seed, p_norm, p_min, p_max, alpha,
    dist_func, use_phase, max_channels, min_len, X_len, prime_scheme
):
    """
    Given a time series dataset and parameters of the method, generate the
    set of random shapelet that will be used in the rest of the algorithm.

    Parameters
    ----------
    X : array, shape=(n_samples, n_features, n_timestamps)
        Time series dataset
    y : array, shape=(n_samples)
        Class of each input time series
    n_shapelets : int
        Number of shapelet to generate
    shapelet_sizes : array, shape=()
        An array of possible shapelet length.
    seed : int
        Random seed generator for numpy
    p_norm : float
        Probability of each shapelet to use z-normalized distance
    p_min : float
        Lower bound for the percentile during the choice of threshold
    p_max : float
        Upper bound for the percentile during the choice of threshold
    alpha : float
        Alpha similarity parameter
    dist_func: function
        A distance function implemented with Numba taking two 1D vectors as
        input.
    use_phase: bool
        Wheter to use phase invariance
    min_len : int
        Minimum length for input time series
    X_len : array, shape=(n_samples)
        The length of each input time series
    prime_scheme : bool
        Wheter to only consider prime numbers as possible dilation 

    Returns
    -------
    set of array, shape=(5)
        values : array, shape=(n_shapelets, max(shapelet_sizes))
            Values of the shapelets. If the shapelet use z-normalized distance,
            those values are already z-normalized by the shapelet 
            initialization step.
        lengths : array, shape=(n_shapelets)
            Length parameter of the shapelets
        dilations : array, shape=(n_shapelets)
            Dilation parameter of the shapelets
        threshold : array, shape=(n_shapelets)
            Threshold parameter of the shapelets
        normalize : array, shape=(n_shapelets)
            Normalization indicatorr of the shapelets
    """
    max_len = max(X_len)
    n_samples, n_features, _ = X.shape
    # Fix the random seed
    seed(r_seed)

    #Initialize shapelets
    values, lengths, dilations, threshold, normalize, n_channels, channel_ids = \
    M_VL_init_random_shapelet_params(
        n_shapelets, shapelet_sizes, min_len, p_norm, max_channels, prime_scheme
    )
    #Initialize self similarity mask
    unique_dil = unique(dilations)
    mask_sampling = ones(
        (2,unique_dil.shape[0],n_samples,n_features,max_len), dtype=bool_
    )
    mask_return = ones(n_shapelets, dtype=bool_)
    #Counter for values array indexes
    a1 = 0
    #Counter for channels_ids array indexes
    a2 = 0     
    for i in prange(n_samples):
        mask_sampling[:,:,i,:,X_len[i]:] = 0
    
    #For each dilation, we can do in parallel
    for i_d in prange(unique_dil.shape[0]):
        #For each shapelet id with this dilation
        id_shps = where(dilations==unique_dil[i_d])[0]
        min_l = min(lengths[id_shps])
        for i_shp in id_shps:
            _dilation = dilations[i_shp]
            _length = lengths[i_shp]
            norm = int64(normalize[i_shp])
            _n_channels = n_channels[i_shp]
            
            _channel_ids = choice(
                arange(0, n_features), _n_channels, replace=False
            )
            
            _values = zeros(_n_channels * _length)
            
            
            mask_dil = mask_sampling[norm,i_d]
            
            i_mask = zeros(n_samples)
            # TODO : the choice of sample don't have the same probability
            # compared to same length version, evaluate the impact.
            if use_phase:
                for i_x in range(n_samples):
                    i_mask[i_x] = _any((
                        mask_dil[i_x][_channel_ids, :X_len[i_x]]
                    ).sum(axis=0)>=_n_channels*alpha)
            else:
                for i_x in range(n_samples):
                    i_mask[i_x] = _any((
                        mask_dil[i_x][_channel_ids, :X_len[i_x]-(_length-1)*_dilation]
                    ).sum(axis=0)>=_n_channels*alpha)
                
            i_mask = where(i_mask)
            if i_mask[0].shape[0] > 0:
                
                #Choose a sample
                id_sample = choice(i_mask[0])
                #Choose a timestamp
                loc_others = where(y == y[id_sample])[0]
                if loc_others.shape[0] > 1:
                    loc_others = loc_others[loc_others != id_sample]
                    id_test = choice(loc_others)
                else:
                    id_test = id_sample
                
                if use_phase:
                    x_dist = zeros(X_len[id_test])
                    t_mask = where(
                        mask_dil[id_sample][
                            _channel_ids ,:X_len[id_sample]
                        ].sum(axis=0)>=_n_channels*alpha
                    )
                else:
                    x_dist = zeros(X_len[id_test]-(_length-1)*_dilation)
                    t_mask = where(
                        mask_dil[id_sample][
                            _channel_ids ,:X_len[id_sample]-(_length-1)*_dilation
                        ].sum(axis=0)>=_n_channels*alpha
                    )
                index = choice(t_mask[0])    
                
                #Counter to keep track of indexes for value affectation
                a3 = 0
                #Select another sample of the same class as the sample used

                
                #Update the mask
                alpha_size = _length - int64(max(1,(1-alpha)*min_l))
                for k in range(_n_channels):
                    for j in range(alpha_size):
                        #We can use modulo event without phase invariance, as we
                        #limit the sampling to d_shape
                        mask_sampling[
                            norm, i_d, id_sample, _channel_ids[k],
                            (index-(j*_dilation))%X_len[id_sample]
                        ] = False
                        mask_sampling[
                            norm, i_d, id_sample, _channel_ids[k],
                            (index+(j*_dilation))%X_len[id_sample]
                        ] = False
                    
                    b3 = a3 + _length
                    #Extract the values
                    _v = get_subsequence(
                        X[id_sample, _channel_ids[k], :X_len[id_sample]], index, _length,
                        _dilation, norm, use_phase
                    )
                    
                    #Compute distance vector
                    x_dist += compute_shapelet_dist_vector(
                        X[id_test, _channel_ids[k], :X_len[id_test]], _v, _length, _dilation,
                        dist_func, norm, use_phase
                    )
                    
                    _values[a3:b3] = _v
                    a3 = b3
                #Counter for values array indexes
                b1 = a1 + _n_channels*_length
                #Counter for channels_ids array indexes
                b2 = a2 + _n_channels
                
                values[a1:b1] = _values
                channel_ids[a2:b2] = _channel_ids
                
                #Extract value between two percentile as threshold for SO
                ps = percentile(x_dist, [p_min,p_max])
                threshold[i_shp] = uniform(
                    ps[0], ps[1]
                )
                a1 = b1
                a2 = b2
            else:
                mask_return[i_shp] = False
            
    return (
        values[:a1],
        lengths[mask_return],
        dilations[mask_return],
        threshold[mask_return],
        normalize[mask_return],
        n_channels[mask_return],
        channel_ids[:a2]
    )

@njit(cache=__USE_NUMBA_CACHE__, parallel=__USE_NUMBA_PARALLEL__, fastmath=__USE_NUMBA_FASTMATH__, nogil=__USE_NUMBA_NOGIL__)
def M_VL_apply_all_shapelets(
    X, shapelets, dist_func, use_phase, X_len
):
    """
    Apply a set of generated shapelet using the parameter arrays previously 
    generated to a set of time series.

    Parameters
    ----------
    X : array, shape=(n_samples, n_features, n_timestamps)
        Input time series
    shapelets: set of array, shape=(5)
        values : array, shape=(n_shapelets, max(shapelet_sizes))
            Values of the shapelets. If the shapelet use z-normalized distance,
            those values are already z-normalized by the shapelet 
            initialization step.
        lengths : array, shape=(n_shapelets)
            Length parameter of the shapelets
        dilations : array, shape=(n_shapelets)
            Dilation parameter of the shapelets
        threshold : array, shape=(n_shapelets)
            Threshold parameter of the shapelets
        normalize : array, shape=(n_shapelets)
            Normalization indicatorr of the shapelets
    dist_func: function
        A distance function implemented with Numba taking two 1D vectors as
        input.
    use_phase: bool
        Wheter to use phase invariance
    X_len : array, shape=(n_samples)
        The length of each input time series
    
    Returns
    -------
    X_new : array, shape=(n_samples, 3*n_shapelets)
        The transformed input time series with each shapelet extracting 3
        feature from the distance vector computed on each time series.

    """
    (values, lengths, dilations, threshold, 
     normalize, n_channels, channel_ids) = shapelets
    n_shapelets = len(lengths)
    n_samples, n_ft, _ = X.shape
    n_features = 3
    
    #(u_l * u_d , 2)
    params_shp = _combinations_1d(lengths, dilations)
    #(u_l * u_d) + 1
    n_shp_params = zeros(params_shp.shape[0]+1, dtype=int64)
    #(n_shapelets)
    idx_shp = zeros(n_shapelets, dtype=int64)
    
    #Indexes per shapelets for values array
    a1 = concatenate((zeros(1, dtype=int64),cumsum(n_channels*lengths)))
    #Indexes per shapelets for channel_ids array
    a2 = concatenate((zeros(1, dtype=int64),cumsum(n_channels)))
    # Counter for shapelet params array
    a3 = 0
    
    for i in range(params_shp.shape[0]):
        _length = params_shp[i, 0]
        _dilation = params_shp[i, 1]
        
        ix_shapelets = where((lengths == _length) & (dilations == _dilation))[0]
        b = a3 + ix_shapelets.shape[0]
        
        idx_shp[a3:b] = ix_shapelets
        n_shp_params[i+1] = ix_shapelets.shape[0]
        
        a3 = b
    n_shp_params = cumsum(n_shp_params)
    X_new = zeros((n_samples, n_features * n_shapelets))
    for i_sample in prange(n_samples):
        #n_shp_params is a cumsum starting at 0
        for i_shp_param in prange(n_shp_params.shape[0]-1):
            _length = params_shp[i_shp_param, 0]
            _dilation = params_shp[i_shp_param, 1]
            strides = generate_strides_2D(
                X[i_sample, :, :X_len[i_sample]], _length, _dilation, use_phase
            )
            # Indexes of shapelets corresponding to the params of i_shp_param
            _idx_shp = idx_shp[n_shp_params[i_shp_param]:n_shp_params[i_shp_param+1]]
            
            _idx_no_norm = _idx_shp[where(normalize[_idx_shp] == False)[0]]
            for i_idx in range(_idx_no_norm.shape[0]):               
                i_shp = _idx_no_norm[i_idx]
                _channels = channel_ids[a2[i_shp]:a2[i_shp+1]]
                _values = values[a1[i_shp]:a1[i_shp+1]].reshape(
                    n_channels[i_shp], _length
                )

                X_new[i_sample, (n_features * i_shp):(n_features * i_shp + n_features)] = \
                apply_one_shapelet_one_sample_multivariate(
                    strides[_channels], _values, threshold[i_shp], dist_func
                )
                        
            _idx_norm = _idx_shp[where(normalize[_idx_shp] == True)[0]]
            if _idx_norm.shape[0] > 0:
                #n_features
                for i_stride in range(strides.shape[0]):
                    #n_timestamps
                    for j_stride in range(strides.shape[1]):
                      _str = strides[i_stride,j_stride]
                      strides[i_stride,j_stride] = (_str - mean(_str))/(std(_str)+1e-8)
                          
                for i_idx in range(_idx_norm.shape[0]):               
                    i_shp = _idx_norm[i_idx]
                    _channels = channel_ids[a2[i_shp]:a2[i_shp+1]]
                    _values = values[a1[i_shp]:a1[i_shp+1]].reshape(
                        n_channels[i_shp], _length
                    )
                    
                    X_new[i_sample, (n_features * i_shp):(n_features * i_shp + n_features)] = \
                    apply_one_shapelet_one_sample_multivariate(
                        strides[_channels], _values, threshold[i_shp], dist_func
                    )
    return X_new


from convst.transformers._commons import generate_strides_2D, generate_strides_1D
import numpy as np
import pytest

def init_numpy(dims):
    return np.random.random_sample(dims)
    
##########################################
#                                        #
#             Test strides               #
#                                        #
##########################################

@pytest.mark.parametrize("dims, window_size, dilation", [
    ((3, 100), 10, 1),
    ((3, 100), 5, 5),
    ((3, 100), 3, 15)
])
def test_strides_2D(dims, window_size, dilation):
    X = init_numpy(dims)
    X2 = generate_strides_2D(X, window_size, dilation, False)
    assert X2.shape == (dims[0], X.shape[1] - (window_size-1)*dilation, window_size)
    assert np.array_equal(X2[0,0], X[0,[0 + j*dilation for j in range(window_size)]])
   
    
@pytest.mark.parametrize("dims, window_size, dilation", [
    ((100), 10, 1),
    ((100), 5, 5),
    ((100), 3, 15)
])
def test_strides_1D(dims, window_size, dilation):
    X = init_numpy(dims)
    X2 = generate_strides_1D(X, window_size, dilation, False)
    assert X2.shape == (X.shape[0] - (window_size-1)*dilation, window_size)
    assert np.array_equal(X2[0], X[[0 + j*dilation for j in range(window_size)]])
    



# -*- coding: utf-8 -*-

from numba import njit, prange
from numpy import float_, sqrt, zeros, unique, bool_, where, int64, log, array
from numpy.random import rand
###############################################################################
#                                                                             #
#                         DISTANCE FUNCTIONS                                  #
#                                                                             #
###############################################################################

from convst import (
    __USE_NUMBA_CACHE__, __USE_NUMBA_FASTMATH__,
    __USE_NUMBA_NOGIL__
)



@njit(
  fastmath=__USE_NUMBA_FASTMATH__, cache=__USE_NUMBA_CACHE__, nogil=__USE_NUMBA_NOGIL__
)
def euclidean(x, y):
    s = 0
    for i in prange(x.shape[0]):
        s += (x[i]-y[i])**2
    return sqrt(s)

@njit(
  fastmath=__USE_NUMBA_FASTMATH__, cache=__USE_NUMBA_CACHE__, nogil=__USE_NUMBA_NOGIL__
)
def squared_euclidean(x, y):
    s = 0
    for i in prange(x.shape[0]):
        s += (x[i]-y[i])**2
    return s

@njit(
  fastmath=__USE_NUMBA_FASTMATH__, cache=__USE_NUMBA_CACHE__, nogil=__USE_NUMBA_NOGIL__
)
def manhattan(x, y):
    s = 0
    for i in prange(x.shape[0]):
        s += abs(x[i]-y[i])
    return s

###############################################################################
#                                                                             #
#                    SUBSEQUENCE EXTRACTION FUNCTIONS                         #
#                                                                             #
###############################################################################

@njit(
  cache=__USE_NUMBA_CACHE__, nogil=__USE_NUMBA_NOGIL__
)
def generate_strides_1D(X, window_size, dilation, use_phase):
    if use_phase:
        return _generate_strides_1D_phase(X, window_size, dilation) 
    else:
        return _generate_strides_1D(X, window_size, dilation)

@njit(
  cache=__USE_NUMBA_CACHE__, nogil=__USE_NUMBA_NOGIL__
)

def generate_strides_2D(X, window_size, dilation, use_phase):
    if use_phase:
        return _generate_strides_2D_phase(X, window_size, dilation) 
    else:
        return _generate_strides_2D(X, window_size, dilation)


@njit(
  cache=__USE_NUMBA_CACHE__, nogil=__USE_NUMBA_NOGIL__
)
def _generate_strides_1D(X, window_size, dilation):
    """
    Generate strides from the input univariate time series with specified 
    length and dilation parameters.

    Parameters
    ----------
    X : array, shape = (n_timestamps)
        An univariate time series, in a 1 dimensional view.
    window_size : int
        Length of the strides to generate.
    dilation : int
        Dilation parameter to apply when generating the strides.

    Returns
    -------
    array, shape = (n_timestamps-(window_size-1)*dilation, window_size)
        All possible subsequences of length stride_len for each time series.

    """
    n_timestamps = X.shape[0] - (window_size - 1) * dilation
    X_new = zeros((n_timestamps, window_size))
    for i in prange(n_timestamps):
        for j in prange(window_size):
            X_new[i,j] = X[i+(j*dilation)]
    return X_new

@njit(
  cache=__USE_NUMBA_CACHE__, nogil=__USE_NUMBA_NOGIL__
)

def _generate_strides_2D(X, window_size, dilation):
    """
    Generate strides from an ensemble of univariate time series with specified 
    length and dilation parameters.

    Parameters
    ----------
    X : array, shape = (n_samples, n_timestamps)
        An ensemble of univariate time series, in a 2 dimensional view. It can
        also be a multivariate time series as (n_features, n_timestamps).
    window_size : int
        Length of the strides to generate.
    dilation : int
        Dilation parameter to apply when generating the strides.

    Returns
    -------
    array, shape = (n_samples, n_timestamps-(window_size-1)*dilation, window_size)
        All possible subsequences of length stride_len for each time series.
    """
    n_features, n_timestamps = X.shape
    n_timestamps = n_timestamps - (window_size - 1) * dilation
    X_new = zeros((n_features, n_timestamps, window_size))
    for ft in prange(n_features):
        for i in prange(n_timestamps):
            for j in prange(window_size):
                X_new[ft,i,j] = X[ft,i+(j*dilation)]
    return X_new


@njit(
  cache=__USE_NUMBA_CACHE__, nogil=__USE_NUMBA_NOGIL__
)
def _generate_strides_1D_phase(X, window_size, dilation):
    """
    Generate strides from the input univariate time series with specified 
    length and dilation parameters. It uses phase invariance.

    Parameters
    ----------
    X : array, shape = (n_timestamps)
        An univariate time series, in a 1 dimensional view.
    window_size : int
        Length of the strides to generate.
    dilation : int
        Dilation parameter to apply when generating the strides.

    Returns
    -------
    array, shape = (n_timestamps, window_size)
        All possible subsequences of length stride_len for each time series.
    """
    n_timestamps = X.shape[0]
    X_new = zeros((n_timestamps,window_size))
    for i in prange(n_timestamps):
        for j in prange(window_size):
            X_new[i,j] = X[(i+(j*dilation))%n_timestamps]
    return X_new


@njit(
  cache=__USE_NUMBA_CACHE__, nogil=__USE_NUMBA_NOGIL__
)
def _generate_strides_2D_phase(X, window_size, dilation):
    """
    Generate strides from an ensemble of univariate time series with specified 
    length and dilation parameters. It uses phase invariance.

    Parameters
    ----------
    X : array, shape = (n_samples, n_timestamps)
        An ensemble of univariate time series, in a 2 dimensional view. It can
        also be a multivariate time series as (n_features, n_timestamps).
    window_size : int
        Length of the strides to generate.
    dilation : int
        Dilation parameter to apply when generating the strides.

    Returns
    -------
    array, shape = (n_samples, n_timestamps, window_size)
        All possible subsequences of length stride_len for each time series.
    """
    n_features, n_timestamps = X.shape
    X_new = zeros((n_features, n_timestamps, window_size))
    for ft in prange(n_features):
        for i in prange(n_timestamps):
            for j in prange(window_size):
                X_new[ft,i,j] = X[ft,(i+(j*dilation))%n_timestamps]
    return X_new


@njit(
  cache=__USE_NUMBA_CACHE__, nogil=__USE_NUMBA_NOGIL__
)
def get_subsequence(X, index, length, d, normalize, use_phase):
    if use_phase:
        return _get_subsequence_phase(
            X, index, length, d, normalize
        )
    else:
        return _get_subsequence(
            X, index, length, d, normalize
        )


@njit(
  cache=__USE_NUMBA_CACHE__, fastmath=__USE_NUMBA_FASTMATH__, nogil=__USE_NUMBA_NOGIL__
)
def _get_subsequence(X, i_start, length, d, normalize):
    """
    Given a set of length and dilation, fetch a subsequence from an input 
    time series given a starting index.

    Parameters
    ----------
    X : array, shape=(n_timestamps)
        Input time series.
    i_start : int
        A starting index between [0, n_timestamps - (l-1)*d]
    l : int
        Length parameter.
    d : int
        Dilation parameter.
    normalize : bool
        Boolean indicating wheter or not the distance will be z-normalized

    Returns
    -------
    v : array, shape=(length)
        The resulting subsequence.

    """
    v = zeros(length)
    _idx = i_start
    
    for j in prange(length):
        v[j] = X[_idx]
        _idx = _idx + d
    
    if normalize:
        v = (v - v.mean())/(v.std()+1e-8)
    return v


@njit(
  cache=__USE_NUMBA_CACHE__, fastmath=__USE_NUMBA_FASTMATH__, nogil=__USE_NUMBA_NOGIL__
)
def _get_subsequence_phase(X, i_start, length, d, normalize):
    """
    Given a set of length and dilation, fetch a subsequence from an input 
    time series given a starting index. It uses phase invariance.

    Parameters
    ----------
    X : array, shape=(n_timestamps)
        Input time series.
    i_start : int
        A starting index between [0, n_timestamps - (l-1)*d]
    l : int
        Length parameter.
    d : int
        Dilation parameter.
    normalize : bool
        Boolean indicating wheter or not the distance will be z-normalized


    Returns
    -------
    v : array, shape=(length)
        The resulting subsequence.

    """
    n_ts = X.shape[0]
    v = zeros(length)
    _idx = i_start
    
    for j in prange(length):
        v[j] = X[_idx]
        _idx = (_idx+d)%n_ts
    
    if normalize:
        v = (v - v.mean())/(v.std()+1e-8)
    return v

###############################################################################
#                                                                             #
#                       DISTANCE VECTOR COMPUTATION                           #
#                                                                             #
###############################################################################


@njit(
  cache=__USE_NUMBA_CACHE__, nogil=__USE_NUMBA_NOGIL__
)
def compute_shapelet_dist_vector(
    x, values, length, dilation, dist_func, normalize, use_phase
):
    if normalize and use_phase:
        return _compute_shapelet_dist_vector_norm_phase(
            x, values, length, dilation, dist_func
        )
    elif normalize and not use_phase:
        return _compute_shapelet_dist_vector_norm(
            x, values, length, dilation, dist_func
        )
    elif not normalize and use_phase:
        return _compute_shapelet_dist_vector_phase(
            x, values, length, dilation, dist_func
        )
    elif not normalize and not use_phase:
        return _compute_shapelet_dist_vector(
            x, values, length, dilation, dist_func
        )
    else:
        raise ValueError('Wrong parameter for normalize or phase')


@njit(
  cache=__USE_NUMBA_CACHE__, fastmath=__USE_NUMBA_FASTMATH__, nogil=__USE_NUMBA_NOGIL__
)
def _compute_shapelet_dist_vector(x, values, length, dilation, dist_func):
    """
    Compute a shapelet distance vector from an univariate time series 
    and a dilated shapelet. Shapelet should be already normalized if normalizing
    the distance

    Parameters
    ----------
    x : array, shape=(n_timestamps)
        An input time series
    values : array, shape=(length)
        The value array of the shapelet
    length : int
        Length of the shapelet
    dilation : int
        Dilation of the shapelet
    normalize : float
        Boolean converted as float (to avoid if statement) indicating
        wheter or not the distance will be z-normalized

    Returns
    -------
    x_conv : array, shape=(n_timestamps - (length-1) * dilation)
        The resulting distance vector

    """
    c = _generate_strides_1D(x, length, dilation)
    x_conv = zeros(c.shape[0])
    for i in prange(x_conv.shape[0]):
        x_conv[i] = dist_func(c[i], values)
    return x_conv


@njit(
  cache=__USE_NUMBA_CACHE__, fastmath=__USE_NUMBA_FASTMATH__, nogil=__USE_NUMBA_NOGIL__
)
def _compute_shapelet_dist_vector_norm(x, values, length, dilation, dist_func):
    """
    Compute a shapelet distance vector from an univariate time series 
    and a dilated shapelet. Shapelet should be already normalized if normalizing
    the distance.

    Parameters
    ----------
    x : array, shape=(n_timestamps)
        An input time series
    values : array, shape=(max_shapelet_length)
        The value array of the shapelet
    length : int
        Length of the shapelet
    dilation : int
        Dilation of the shapelet
    normalize : float
        Boolean converted as float (to avoid if statement) indicating
        wheter or not the distance will be z-normalized

    Returns
    -------
    x_conv : array, shape=(n_timestamps - (length-1) * dilation)
        The resulting distance vector

    """
    c = _generate_strides_1D(x, length, dilation)
    x_conv = zeros(c.shape[0])
    for i in prange(x_conv.shape[0]):
        x0 = (c[i] - c[i].mean())/(c[i].std()+1e-8)
        x_conv[i] = dist_func(x0, values)
    return x_conv


@njit(
  cache=__USE_NUMBA_CACHE__, fastmath=__USE_NUMBA_FASTMATH__, nogil=__USE_NUMBA_NOGIL__
)
def _compute_shapelet_dist_vector_phase(x, values, length, dilation, dist_func):
    """
    Compute a shapelet distance vector from an univariate time series 
    and a dilated shapelet. Shapelet should be already normalized if normalizing
    the distance. It uses phase invariance.

    Parameters
    ----------
    x : array, shape=(n_timestamps)
        An input time series
    values : array, shape=(length)
        The value array of the shapelet
    length : int
        Length of the shapelet
    dilation : int
        Dilation of the shapelet
    normalize : float
        Boolean converted as float (to avoid if statement) indicating
        wheter or not the distance will be z-normalized

    Returns
    -------
    x_conv : array, shape=(n_timestamps)
        The resulting distance vector

    """
    c = _generate_strides_1D_phase(x, length, dilation)
    x_conv = zeros(c.shape[0])
    for i in prange(x_conv.shape[0]):
        x_conv[i] = dist_func(c[i], values)
    return x_conv


@njit(
  cache=__USE_NUMBA_CACHE__, fastmath=__USE_NUMBA_FASTMATH__, nogil=__USE_NUMBA_NOGIL__
)
def _compute_shapelet_dist_vector_norm_phase(x, values, length, dilation, dist_func):
    """
    Compute a shapelet distance vector from an univariate time series 
    and a dilated shapelet. Shapelet should be already normalized if normalizing
    the distance. It uses phase invariance.

    Parameters
    ----------
    x : array, shape=(n_timestamps)
        An input time series
    values : array, shape=(max_shapelet_length)
        The value array of the shapelet
    length : int
        Length of the shapelet
    dilation : int
        Dilation of the shapelet
    normalize : float
        Boolean converted as float (to avoid if statement) indicating
        wheter or not the distance will be z-normalized

    Returns
    -------
    x_conv : array, shape=(n_timestamps)
        The resulting distance vector

    """
    c = _generate_strides_1D_phase(x, length, dilation)
    x_conv = zeros(c.shape[0])
    for i in prange(x_conv.shape[0]):
        x0 = (c[i] - c[i].mean())/(c[i].std()+1e-8)
        x_conv[i] = dist_func(x0, values)
    return x_conv



@njit(
  cache=__USE_NUMBA_CACHE__, fastmath=__USE_NUMBA_FASTMATH__, nogil=__USE_NUMBA_NOGIL__
)
def apply_one_shapelet_one_sample_univariate(x, values, threshold, dist_func):
    """
    Extract the three features from the distance between a shapelet and the 
    strides of an input time series generated by the length and dilation 
    parameter of the shapelet. If normalization should be used, all strides
    of should be z-normalized independently.

    Parameters
    ----------
    x : array, shape=(n_timestamps - (length-1)*dilation, length)
        Strides of an input time series
    values : array, shape=(max(shapelet_sizes))
        Values of the shapelet
    threshold : float
        The threshold to compute the shapelet occurence feature.
    
    Returns
    -------
    _min : float
        The minimum euclidean distance between the shapelet and the input time
        series
    float
        The location of the minimum euclidean distance divided by the length 
        of the distance vector (i.e scaled between [0,1]).        
    float
        The number of points in the distance vector inferior to the threshold 
        divided by the length of the distance vector (i.e scaled between [0,1])

    """
    n_candidates, length = x.shape

    _n_match = 0
    _min = -1. 
    _argmin = 0

    #For each step of the moving window in the shapelet distance
    for i in prange(n_candidates):
        _dist = dist_func(x[i], values)

        if _dist < _min or _min==-1.:
            _min = _dist
            _argmin = i

        if _dist <= threshold:
            _n_match += 1
            
    return _min, float_(_argmin), float_(_n_match)


@njit(
  cache=__USE_NUMBA_CACHE__, fastmath=__USE_NUMBA_FASTMATH__, nogil=__USE_NUMBA_NOGIL__
)
def apply_one_shapelet_one_sample_multivariate(x, values, threshold, dist_func):
    """
    Extract the three features from the distance between a shapelet and the 
    strides of an input time series generated by the length and dilation 
    parameter of the shapelet. If normalization should be used, all strides
    of should be z-normalized independently.

    Parameters
    ----------
    x : array, shape=(n_timestamps - (length-1)*dilation, n_features, length)
        Strides of an input time series
    values : array, shape=(max(shapelet_sizes))
        Values of the shapelet
    threshold : float
        The threshold to compute the shapelet occurence feature.
    
    Returns
    -------
    _min : float
        The minimum euclidean distance between the shapelet and the input time
        series
    float
        The location of the minimum euclidean distance divided by the length 
        of the distance vector (i.e scaled between [0,1]).        
    float
        The number of points in the distance vector inferior to the threshold 
        divided by the length of the distance vector (i.e scaled between [0,1])

    """
    n_ft, n_candidates, length = x.shape

    _n_match = 0
    _min = -1.
    _argmin = 0
    
    #For each step of the moving window in the shapelet distance
    for i in prange(n_candidates):
        _dist = 0
        for ft in prange(n_ft):
            _dist += dist_func(x[ft, i], values[ft])
    
        if _dist < _min or _min == -1.:
            _min = _dist
            _argmin = i
            
        if _dist <= threshold:
            _n_match += 1
    
    return _min, float_(_argmin), float_(_n_match)



@njit(
  cache=__USE_NUMBA_CACHE__, nogil=__USE_NUMBA_NOGIL__
)
def _combinations_1d(x,y):
    """
    Return the unique combination (in the 2nd dimension) of the 2D array made by
    concatenating x and y.

    Parameters
    ----------
    x : array, shape=(u)
        A 1D array of values
    y : array, shape=(v)
        A 1D array of values

    Returns
    -------
    array, shape=(w, 2)
        The unique combination (in the 2nd dimension) of the 2D array made by
        concatenating x and y.

    """
    # Fix issues with multiple length, but could be optimized
    u_x = unique(x)
    u_y = unique(y)
    u_mask = zeros((u_x.shape[0], u_y.shape[0]),dtype=bool_)
    
    for i in range(x.shape[0]):
        u_mask[where(u_x==x[i])[0][0],where(u_y==y[i])[0][0]] = True
    combinations = zeros((u_mask.sum(),2), dtype=int64)
    i_comb = 0
    for i in range(x.shape[0]):
        if u_mask[where(u_x==x[i])[0][0],where(u_y==y[i])[0][0]]:
            combinations[i_comb,0] = x[i]
            combinations[i_comb,1] = y[i]
            u_mask[where(u_x==x[i])[0][0],where(u_y==y[i])[0][0]] = False
            i_comb += 1
    return combinations

@njit(
  cache=__USE_NUMBA_CACHE__, nogil=__USE_NUMBA_NOGIL__
)
def prime_up_to(n):
    is_p = zeros(n+1, dtype=bool_)
    for i in range(n+1):
        is_p[i] = is_prime(i)
    return where(is_p)[0]


@njit(
  cache=__USE_NUMBA_CACHE__, nogil=__USE_NUMBA_NOGIL__
)
def is_prime(n):
    if (n % 2 == 0 and n > 2) or n == 0: 
        return False
    for i in range(3, int64(sqrt(n)) + 1, 2):
        if not n % i:
            return False 
    return True

@njit(
  cache=__USE_NUMBA_CACHE__, fastmath=__USE_NUMBA_FASTMATH__, nogil=__USE_NUMBA_NOGIL__
)
def choice_log(n_choice, n_sample):
    if n_choice > 1:
        P = array([1/2**log(i) for i in range(1,n_choice+1)])
        P = P.cumsum()/P.sum()
        loc = zeros(n_sample, dtype=int64)
        for i in prange(n_sample):
            loc[i] = where(P>=rand())[0][0]
        return loc
    else:
        return zeros(n_sample, dtype=int64)



# -*- coding: utf-8 -*-

from convst.utils.dataset_utils import (
    load_sktime_dataset_split, load_sktime_dataset, z_norm_3D
)
import numpy as np
import pytest


def init_numpy(dims):
    return np.random.random_sample(dims)

"""
@pytest.mark.parametrize("dims", [
    ((30, 15, 12)),
    ((2, 1, 18)),
])
"""

def test_z_norm_3D():
    X0 = np.zeros((10,1,10))
    X1 = np.ones((10,1,10))
    assert np.array_equal(z_norm_3D(X0), X0)
    assert np.array_equal(z_norm_3D(X1), X0)
    
    
@pytest.mark.parametrize("name", [
    ('GunPoint'), ('SmoothSubspace'),
])
def test_load_sktime_dataset_split(name):
    X_train, X_test, y_train, y_test, le = load_sktime_dataset_split(
        name=name, normalize=False
    )
    if name == 'GunPoint':
        assert X_train.shape == (50,1,150)
        assert X_test.shape == (150,1,150)
    elif name == 'SmoothSubspace':
        assert X_train.shape == (150,1,15)
        assert X_test.shape == (150,1,15)
        
    assert y_train.shape[0] == X_train.shape[0]
    assert y_test.shape[0] == X_test.shape[0]

@pytest.mark.parametrize("name", [
    ('GunPoint'), ('SmoothSubspace'),
])
def test_load_sktime_dataset(name):
    X, y, le = load_sktime_dataset(
        name=name, normalize=False
    )
    if name == 'GunPoint':
        assert X.shape == (200,1,150)
    elif name == 'SmoothSubspace':
        assert X.shape == (300,1,15)
    assert y.shape[0] == X.shape[0]


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 26 12:25:16 2022
"""

# TODO: Define class taking as input a RDST transformer,
# giving option to visualize each shapelets as base class

# Then, define interpreter for RDST Ridge

# And for RDST Ensemble, using RDST Ridge interpreter.


import seaborn as sns
from matplotlib import pyplot as plt
import numpy as np
from sklearn.utils.validation import check_is_fitted

from convst.transformers._commons import compute_shapelet_dist_vector, manhattan
from convst.transformers import R_DST
from convst.classifiers import R_DST_Ridge, R_DST_Ensemble

class Shapelet:
    """
    A Shapelet object to use for ploting operations

    Parameters
    ----------
    values : array, shape=(n_channels * length)
        Values of the shapelet
    length : int
        Length of the shapelet
    dilation : int
        Dilation of the shapelet
    norm : bool
        Does the shapelet use a normalized distance ?
    threshold : float
        Lambda threshold for Shapelet Occurrence feature
    phase : bool
        Does the shapelet use phase invariance ?

    Returns
    -------
    None.

    """
    def __init__(self, values, length, dilation, norm, threshold, phase):
        self.values = np.asarray(values)
        self.length = length
        self.dilation = dilation
        self.norm = norm
        self.phase = phase
        self.threshold = threshold
        
    def plot(self, figsize=(10,5), seaborn_context='talk', ax=None):
        """
        Plot the shapelet values

        Parameters
        ----------
        figsize : tuple, optional
            2D size of the figure. The default is (10,5).
        seaborn_context : str, optional
            Seaborn module context. The default is 'talk'.
        ax : matplotlib axe, optional
            A matplotlib axe on which to plot the figure. The default is None 
            and will create a new figure of size figsize.

        Returns
        -------
        fig : matplotlib figure
            The resulting figure 
        """
        if ax is None:
            sns.set()
            sns.set_context(seaborn_context)
            fig = plt.figure(figsize=(figsize))
            plt.plot(self.values)
            plt.title(
                'd={},normalize={},threshold={}'.format(
                    self.dilation, self.norm, np.round(self.threshold,decimals=2)
                )
            )
            return fig
        else:
            ax.plot(self.values)
            ax.set_title(
                'd={},normalize={},threshold={}'.format(
                    self.dilation, self.norm, np.round(self.threshold,decimals=2)
                )
            )
            return ax
        
    def plot_on_X(
        self, X, d_func=manhattan, figsize=(10,5), seaborn_context='talk', alpha=0.9,
        shp_dot_size=40, shp_c='purple', ax=None, label=None, x_linewidth=2
    ):
        """
        Plot the shapelet on its best match on the time series X

        Parameters
        ----------
        X : array, shape=(n_timestamps) or shape=(n_features, n_timestamps)
            Input time series
        d_func : object, optional
            A distance function between two arrays. The default is manhattan.
        figsize : tuple, optional
            Size of the figure. The default is (10,5).
        seaborn_context : str, optional
            Seaborn context. The default is 'talk'.
        alpha : float, optional
            Alpha parameter for plotting X. The default is 0.9.
        shp_dot_size : float, optional
            Size of the scatter plot to represent the shapelet on X.
            The default is 40.
        shp_c : str, optional
            Color of the shapelet scatter plot. The default is 'purple'.
        ax : matplotlib axe, optional
            A matplotlib axe on which to plot the figure. The default is None 
            and will create a new figure of size figsize.
        label : str, optional
            Custom label to plot as legend for X. The default is None.
        x_linewidth : float, optional
            The linewidth of X plot. The default is 2.

        Returns
        -------
        fig : matplotlib figure
            The resulting figure with S on its best match on X. A normalized
            shapelet will be scalled to macth the scale of X.

        """
        c = compute_shapelet_dist_vector(
            X, self.values, self.length, self.dilation,
            manhattan, self.norm, self.phase
        )
        _values = self.values
        idx_match = np.asarray(
            [(c.argmin() + i*self.dilation)%X.shape[0] for i in range(self.length)]
        ).astype(int)
        if self.norm:
            _values = (_values * X[idx_match].std()) + X[idx_match].mean()
            
        if ax is None:
            sns.set()
            sns.set_context(seaborn_context)
            fig = plt.figure(figsize=(figsize))
            plt.plot(X,label=label, linewidth=x_linewidth, alpha=alpha)
            plt.scatter(idx_match, _values, s=shp_dot_size, c=shp_c, zorder=3, alpha=alpha)
            return fig
        else:
            ax.plot(X,label=label, linewidth=x_linewidth, alpha=alpha)
            ax.scatter(idx_match, _values, s=shp_dot_size, c=shp_c, zorder=3, alpha=alpha)
    
    def plot_distance_vector(   
        self, X, d_func=manhattan, figsize=(10,5), seaborn_context='talk',
        c_threshold='purple', ax=None, label=None
    ):
        """
        Plot the shapelet distance vector computed between itself and X.

        Parameters
        ----------
        X : array, shape=(n_timestamps) or shape=(n_features, n_timestamps)
            Input time series
        d_func : object, optional
            A distance function between two arrays. The default is manhattan.
        figsize : tuple, optional
            Size of the figure. The default is (10,5).
        seaborn_context : str, optional
            Seaborn context. The default is 'talk'.
        c_threshold : float, optional
            Color used to represent a line on the y-axis to visualize the lambda
            threshold. The default is 'purple'.
        ax : matplotlib axe, optional
            A matplotlib axe on which to plot the figure. The default is None 
            and will create a new figure of size figsize.
        label : str, optional
            Custom label to plot as legend. The default is None.
        
        Returns
        -------
        fig : matplotlib figure
            The resulting figure with the distance vector obtained by d(S,X)

        """
        c = compute_shapelet_dist_vector(
            X, self.values, self.length, self.dilation,
            manhattan, self.norm, self.phase
        )
        if ax is None:
            sns.set()
            sns.set_context(seaborn_context)
            fig = plt.figure(figsize=(figsize))
            plt.plot(c,label=label)
            plt.hlines(self.threshold, 0, c.shape[0], color=c_threshold)
            return fig
        else:
            ax.plot(c,label=label)
            ax.hlines(self.threshold, 0, c.shape[0], color=c_threshold)
            return ax

    
class RDST_interpreter():
    """
    A class to interpret the result from a fitted RDST instance.    

    Parameters
    ----------
    RDST : object
        A fitted RDST transformer.

    Returns
    -------
    None.

    """
    
    def __init__(self, RDST):
        check_is_fitted(RDST, ['shapelets_'])
        if isinstance(RDST, R_DST):
            self.RDST = RDST
        else:
            raise TypeError(
                'Object passed to RDST interpreter should be an R_DST instance'
            )
        
    def _get_params(self, id_shapelet):
        values, lengths, dilations, threshold, normalize = self.RDST.shapelets_
        phase = self.RDST.phase_invariance
        if self.RDST.transform_type in ['multivariate','multivariate_variable']:
            raise NotImplementedError('Interpreter is not yet implemented for multivariate data')
        else:
            length_ = lengths[id_shapelet]
            values_ = values[id_shapelet, :length_]
            dilation = dilations[id_shapelet]
            norm = normalize[id_shapelet]
            threshold_ = threshold[id_shapelet]
        return values_, length_, dilation, norm, threshold_, phase 
        
        
    def plot_on_X(
        self, id_shapelet, X, d_func=manhattan, figsize=(10,5),
        seaborn_context='talk', shp_dot_size=40, shp_c='purple', ax=None,
        label=None
    ):
        """
        Plot the shapelet on its best match on the time series X

        Parameters
        ----------
        id_shapelet : int
            ID of the shapelet to plot.
        X : array, shape=(n_timestamps) or shape=(n_features, n_timestamps)
            Input time series
        d_func : object, optional
            A distance function between two arrays. The default is manhattan.
        figsize : tuple, optional
            Size of the figure. The default is (10,5).
        seaborn_context : str, optional
            Seaborn context. The default is 'talk'.
        alpha : float, optional
            Alpha parameter for plotting X. The default is 0.9.
        shp_dot_size : float, optional
            Size of the scatter plot to represent the shapelet on X.
            The default is 40.
        shp_c : str, optional
            Color of the shapelet scatter plot. The default is 'purple'.
        ax : matplotlib axe, optional
            A matplotlib axe on which to plot the figure. The default is None 
            and will create a new figure of size figsize.
        label : str, optional
            Custom label to plot as legend for X. The default is None.
        x_linewidth : float, optional
            The linewidth of X plot. The default is 2.

        Returns
        -------
        fig : matplotlib figure
            The resulting figure with S on its best match on X. A normalized
            shapelet will be scalled to macth the scale of X.

        """
        return Shapelet(*self._get_params(id_shapelet)).plot_on_X(
            X, d_func=d_func, figsize=figsize,
            seaborn_context=seaborn_context, shp_dot_size=shp_dot_size,
            shp_c=shp_c, ax=ax, label=label
        )

    def plot_distance_vector(
        self, id_shapelet, X, d_func=manhattan, figsize=(10,5), 
        seaborn_context='talk', c_threshold='purple', ax=None, label=None
    ):
        """
        Plot the shapelet distance vector computed between itself and X.

        Parameters
        ----------
        id_shapelet : int
            ID of the shapelet to plot.
        X : array, shape=(n_timestamps) or shape=(n_features, n_timestamps)
            Input time series
        d_func : object, optional
            A distance function between two arrays. The default is manhattan.
        figsize : tuple, optional
            Size of the figure. The default is (10,5).
        seaborn_context : str, optional
            Seaborn context. The default is 'talk'.
        c_threshold : float, optional
            Color used to represent a line on the y-axis to visualize the lambda
            threshold. The default is 'purple'.
        ax : matplotlib axe, optional
            A matplotlib axe on which to plot the figure. The default is None 
            and will create a new figure of size figsize.
        label : str, optional
            Custom label to plot as legend. The default is None.
        
        Returns
        -------
        fig : matplotlib figure
            The resulting figure with the distance vector obtained by d(S,X)

        """

        return Shapelet(*self._get_params(id_shapelet)).plot_distance_vector(
            X, d_func=d_func, figsize=figsize, seaborn_context=seaborn_context,
            c_threshold=c_threshold, ax=ax, label=label
        )
    
    def plot(self, id_shapelet, figsize=(10,5), seaborn_context='talk', ax=None):
        """
        Plot the shapelet values

        Parameters
        ----------
        id_shapelet : int
            ID of the shapelet to plot.
        figsize : tuple, optional
            2D size of the figure. The default is (10,5).
        seaborn_context : str, optional
            Seaborn module context. The default is 'talk'.
        ax : matplotlib axe, optional
            A matplotlib axe on which to plot the figure. The default is None 
            and will create a new figure of size figsize.

        Returns
        -------
        fig : matplotlib figure
            The resulting figure 
        """
        return Shapelet(*self._get_params(id_shapelet)).plot(
            figsize=figsize, seaborn_context=seaborn_context, ax=ax
        )
    

class RDST_Ridge_interpreter():
    """
    Interpreter for RDST Ridge models.

    Parameters
    ----------
    RDST_Ridge : object
        A fitted RDST Ridge classifier.

    Returns
    -------
    None.

    """
    
    def __init__(self, RDST_Ridge):
        check_is_fitted(RDST_Ridge, ['classifier'])
        if isinstance(RDST_Ridge, R_DST_Ridge):
            self.RDST_Ridge = RDST_Ridge
        else:
            raise TypeError(
                'Object passed to RDST_Ridge interpreter should be an R_DST_Ridge instance'
            )
        self.rdst_interp = RDST_interpreter(self.RDST_Ridge.transformer)
        
    def _get_shp_importance(self, class_id):
        coefs = self.RDST_Ridge.classifier['ridgeclassifiercv'].coef_
        n_classes = coefs.shape[0]
        
        if n_classes == 1:
            coefs = np.append(-coefs, coefs, axis=0)
        c_ = np.zeros(self.RDST_Ridge.transformer.shapelets_[1].shape[0]*3)
        c_[self.RDST_Ridge.classifier['c_standardscaler'].usefull_atts] = coefs[class_id]
        return c_
    
    def visualize_best_shapelets_one_class(
        self, X, y, class_id, n_shp=1, figsize=(16,12), seaborn_context='talk'
    ):
        """
        Plot the n_shp best candidates for the class_id. Visualize best macth
        on two random samples and how the shapelet discriminate (X,y) with boxplots

        Parameters
        ----------
        X : array, shape=(n_samples, n_fetaures, n_timestamps)
            A time series dataset. Can be the training set to visualize training
            results, or testing to visualize generalization to unseen samples.
        y : array, shape=(n_samples)
            The true classes of the time series dataset.
        class_id : int
            ID of the class we want to visualize. The n_shp best shapelet for
            this class will be selected based on the feature coefficients
            inside the ridge classifier.
        n_shp : int, optional
            Number of plots to output, one per shapelet (i.e. the n_shp best shapelets for class_id).
            The default is 1.
        figsize : tuple, optional
            Size of the figure. The default is (16,12).
        seaborn_context : str, optional
            Seaborn context. The default is 'talk'.
           
        Returns
        -------
        None.

        """
        sns.set()
        sns.set_context(seaborn_context)
        
        coefs = self._get_shp_importance(class_id)
        
        idx = (coefs.argsort()//3)[::-1]
        shp_ids = []
        i=0
        while len(shp_ids)<n_shp and i < idx.shape[0]:
            if idx[i] not in shp_ids:
                shp_ids = shp_ids + [idx[i]]
            i+=1
            
        X_new = self.RDST_Ridge.transformer.transform(X)
        i_example = np.random.choice(np.where(y==class_id)[0])
        i_example2 = np.random.choice(np.where(y!=class_id)[0])
        y_copy = (y == class_id).astype(int)
        for i_shp in shp_ids:
            fig, ax = plt.subplots(nrows=2, ncols=3, figsize=figsize)
            
            ax[0,0].set_title('Boxplot of min')
            sns.boxplot(x=y_copy,y=X_new[:,(i_shp*3)],ax=ax[0,0])
            ax[0,0].set_xticklabels(['Other classes', 'Class {}'.format(class_id)])
            
            ax[0,1].set_title('Boxplot of argmin')
            sns.boxplot(x=y_copy,y=X_new[:,1+(i_shp*3)],ax=ax[0,1])
            
            ax[0,1].set_xticklabels(['Other classes', 'Class {}'.format(class_id)])
            ax[0,2].set_title('Boxplot of Shapelet Occurence')
            sns.boxplot(x=y_copy,y=X_new[:,2+(i_shp*3)],ax=ax[0,2])
            
            ax[0,2].set_xticklabels(['Other classes', 'Class {}'.format(class_id)])            
            
            ax[1,0].set_title('Best match')
            ax[1,2].set_title('Distance vectors')
            
            
            self.rdst_interp.plot(i_shp, ax=ax[1,1])
            
            self.rdst_interp.plot_on_X(i_shp, X[i_example2,0], ax=ax[1,0], label='Other class')
            self.rdst_interp.plot_on_X(i_shp, X[i_example,0], ax=ax[1,0], label='Class {}'.format(class_id))
            
            self.rdst_interp.plot_distance_vector(i_shp, X[i_example2,0], ax=ax[1,2], label='Other class')
            self.rdst_interp.plot_distance_vector(i_shp, X[i_example,0], ax=ax[1,2], label='Class {}'.format(class_id))
        
            ax[1,0].legend()
            plt.show()

            
class RDST_Ensemble_interpreter:
    def __init__(self, RDST_Ensemble):
        check_is_fitted(RDST_Ensemble, ['classifier'])
        if isinstance(RDST_Ensemble, R_DST_Ensemble):
            self.RDST_Ensemble = RDST_Ensemble
        else:
            raise TypeError(
                'Object passed to RDST_Ridge interpreter should be an R_DST_Ridge instance'
            )
    
    def visualize_best_shapelets_one_class(self, class_id, n_shp=1):
        raise NotImplementedError()


"""
This module offers sklearn compatible classifiers which direclty wrap R_DST with some classifiers. They are equivalent to making a pipeline but are shorter to write. 
"""
from .rdst_ridge import R_DST_Ridge
from .rdst_ensemble import R_DST_Ensemble

__author__ = 'Antoine Guillaume antoine.guillaume45@gmail.com'

__all__ = ["R_DST_Ridge","R_DST_Ensemble"]


# -*- coding: utf-8 -*-

import numpy as np

from joblib import Parallel

from convst.utils.checks_utils import (check_n_jobs, check_array_1D,
                                       check_is_boolean, check_is_numeric)
from convst.transformers import R_DST
from convst.transformers._input_transformers import (
    Raw, Derivate, Periodigram
)
from sklearn.utils.fixes import delayed
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.linear_model import RidgeClassifierCV
from sklearn.utils.extmath import softmax
from sklearn.metrics import accuracy_score, make_scorer
from sklearn.preprocessing import StandardScaler
from sklearn.pipeline import make_pipeline

from sklearn.utils.validation import check_is_fitted, check_random_state

from numba import set_num_threads

class _internalRidgeCV(RidgeClassifierCV): 
    def __init__(self, **kwargs):
        super().__init__(
            store_cv_values=True,
            scoring=make_scorer(accuracy_score),
            **kwargs
        )
    
    def fit(self, X, y):
        self.scaler = StandardScaler().fit(X)
        return super().fit(self.scaler.transform(X), y)
    
    def predict(self, X):
        return super().predict(self.scaler.transform(X))
    
    def predict_proba(self, X):
        d = self.decision_function(self.scaler.transform(X))
        if len(d.shape) == 1:
            d = np.c_[-d, d]
        return softmax(d)
    
    def get_loocv_train_acc(self, y_train):
        if self.store_cv_values:
            alpha_idx = np.where(self.alphas == self.alpha_)[0]
            cv_vals = self.cv_values_[:,:,alpha_idx]
            if cv_vals.shape[1] == 1:
                return accuracy_score((cv_vals[:,0]>0).astype(int), y_train)
            else:
                return accuracy_score(cv_vals.argmax(axis=1), y_train)
        else:
            raise ValueError('LOOCV training accuracy is only available with store_cv_values to True')


def _parallel_fit(X, y, model):
    return model.fit(X, y)

def _parallel_predict(X, model, w):
    return model.predict_proba(X) * w

class R_DST_Ensemble(BaseEstimator, ClassifierMixin):
    """
    
    Parameters
    ----------
    transform_type : str, optional
        Type of transformer to use. Based on the characteristics of the input
        time series, different class of transformer must be used, for example
        the tranformer for univariate series is not the same as for
        multivariate ones for run-time optimization reasons.
        The default is 'auto', which automatically select the transformer based
        on the data passed in the fit method.
    phase_invariance : bool, optional
        Wheter to use phase invariance for shapelet sampling and distance 
        computation. The default is False.
    distance : str, optional
        The distance function to use whe computing distances between shapelets
        and time series. Choose between 'euclidean','manhattan' and 'squared_euclidean'.
        The default is 'manhattan'.
    alpha : float, optional
        The alpha similarity parameter, the higher the value, the lower the 
        allowed number of common indexes with previously sampled shapelets 
        when sampling a new one with similar parameters. It can cause the
        number of sampled shapelets to be lower than n_shapelets if the
        whole search space has been covered. The default is 0.5.
    normalize_output : boolean, optional
        Wheter to normalize the argmin and shapelet occurrence feature by the 
        length of the series from which it was extracted. This is mostly useful
        for variable length time series. The default is False.
    n_samples : float, optional
        Proportion (in ]0,1]) of samples to consider for the shapelet
        extraction. The default is None, meaning that all samples are used.
    n_shapelets : int, optional
        The maximum number of shapelet to be sampled. The default is 10_000.
    shapelet_lengths : array, optional
        The set of possible length for shapelets. The values can be integers
        to specify an absolute length, or a float, to specify a length relative 
        to the input time series length. The default is [11].
    shapelet_lengths_bounds : array, optional
        An 1D array with two elements containing the min and max possible 
        length for shapelet candidate, can be int or float. The default is
        None, meaning that shapelet_lengths parameter is used.
    lengths_bounds_reduction : float, optional
        A float in ]0,1], quantifying the proportion of lengths to explore 
        between the min and max bounds of shapelet_lengths_bounds. The default
        is 0.5. For example, with bounds as [4,10], and a reduction of 0.5,
        only [4,6,8,10] will be considered as possible lengths.
    prime_dilations : bool, optional
        If True, only dilation with prime values will be considered for 
        shapelet candidates. This will greatly speed-up the algorithm
        for long time series and/or short shapelet length, possibly at the cost
        of some accuracy.
    proba_norm : float, optional
        The proportion of shapelets that will use a normalized distance 
        function, which induce scale invariance. The default is 0.8.
    percentiles : array, optional
        The two perceniles used to select the lambda threshold used to compute
        the Shapelet Occurrence feature. The default is [5,10].
    n_jobs : int, optional
        The number of threads used to sample and compute the distance vectors.
        The default is 1, -1 means all available cores.
    random_state : object, optional
        The seed for the random state. The default is None.
    max_channels : int, optional
        The maximum number of feature possibly considered by a multivariate 
        shapelet. The default is None, meaning max_chanels=n_features.
    min_len : int, optional
        The minimum length of an input time series for variable length input.
        The default is None, meaning min_len=min(n_timestamps) on the training data.
        This can cause error if a shorter serie sis present in the test set.
    class_weight : object, optional
         Class weight option of Ridge Classifier, either None, "balanced" or a
         custom dictionnary of weight for each class. The default is None.
    fit_intercept : bool, optional
         If True, the intercept term will be fitted during the ridge regression.
         The default is True.
    alphas : array, optional
         Array of alpha values to try which influence regularization strength, 
         must be a positive float. The default is np.logspace(-4,4,20).
    """
    def __init__(
        self,
        n_shapelets_per_estimator=10000,
        shapelet_lengths=[11],
        shapelet_lengths_bounds=None,
        lengths_bounds_reduction=0.5,
        prime_dilations=False,
        n_samples=None,
        n_jobs=1,
        prefer=None,
        require='sharedmem',
        random_state=None,
        shp_alpha=0.5,
        a_w=4,
        proba_norm=[0.8, 0.8, 0.8],
        phase_invariance=False,
        input_transformers=None,
        transform_type='auto',
        distance='manhattan',
        normalize_output=False,
        percentiles=[5,10],
        max_channels=None,
        min_len=None,
        class_weight=None,
        fit_intercept=True,
        alphas_ridge=list(np.logspace(-4,4,20))
    ):
        self.transform_type = transform_type
        self.phase_invariance = check_is_boolean(phase_invariance)
        self.distance = distance
        self.normalize_output = check_is_boolean(normalize_output)
        self.n_samples = check_is_numeric(n_samples) if n_samples is not None else n_samples
        self.shapelet_lengths_bounds = shapelet_lengths_bounds
        self.prime_dilations = check_is_boolean(prime_dilations)
        self.percentiles = percentiles
        self.random_state = check_random_state(random_state)
        self.max_channels=max_channels
        self.min_len=min_len
        self.lengths_bounds_reduction = lengths_bounds_reduction
        self.n_shapelets_per_estimator = check_is_numeric(n_shapelets_per_estimator)
        self.shapelet_lengths = check_array_1D(shapelet_lengths)
        self.n_jobs = n_jobs
        
        self.prefer = prefer
        self.require = require
        self.shp_alpha = check_is_numeric(shp_alpha)
        self.a_w = check_is_numeric(a_w)
        self.proba_norm = check_array_1D(proba_norm)
        
        self.alphas_ridge = alphas_ridge
        self.class_weight = class_weight
        self.fit_intercept = fit_intercept
      
        if input_transformers is None:
            self.input_transformers = [
                Raw(),
                Derivate(),
                Periodigram()
            ]
        else:
            self.input_transformers = input_transformers
        
        if len(self.input_transformers) != len(self.proba_norm):
            raise ValueError(
                'The length of proba norm array should be equal to the'
                ' length of the input transformers array, but found '
                '{} for proba_norm and {} for input transformers'.format(
                 len(self.proba_norm), len(self.input_transformers))
            )
        
    def _more_tags(self):
        return {
            "capability:variable_length": True,
            "capability:univariate": True,
            "capability:multivariate": True
        }
        
    def _manage_n_jobs(self):
        total_jobs = check_n_jobs(self.n_jobs)
        self.n_jobs = min(len(self.input_transformers),total_jobs)
        self.n_jobs_rdst = max(1,total_jobs//self.n_jobs)
        
    def fit(self, X, y):
        self._manage_n_jobs()
        set_num_threads(self.n_jobs_rdst)
        models = Parallel(
            n_jobs=self.n_jobs,
            prefer=self.prefer,
            require=self.require
        )(
            delayed(_parallel_fit)(
                X, y, 
                make_pipeline(
                    self.input_transformers[i],
                    R_DST(
                        n_shapelets=self.n_shapelets_per_estimator,
                        alpha=self.shp_alpha, n_samples=self.n_samples, 
                        proba_norm=self.proba_norm[i],
                        n_jobs=False,
                        shapelet_lengths=self.shapelet_lengths,
                        phase_invariance=self.phase_invariance,
                        prime_dilations=self.prime_dilations,
                        shapelet_lengths_bounds=self.shapelet_lengths_bounds,
                        lengths_bounds_reduction=self.lengths_bounds_reduction,
                        transform_type=self.transform_type,
                        percentiles=self.percentiles,
                        min_len=self.min_len,
                        max_channels=self.max_channels,
                        random_state=self.random_state
                    ),
                    _internalRidgeCV(
                        alphas=self.alphas_ridge,
                        fit_intercept=self.fit_intercept,
                        class_weight=self.class_weight
                    )
                )
            )
            for i in range(len(self.input_transformers))
        )
            
        self.models = models
        self.model_weights = [model['_internalridgecv'].get_loocv_train_acc(y)**self.a_w
                              for model in models]
        return self


    def predict(self, X):
        check_is_fitted(self, ['model_weights'])
        preds_proba = Parallel(
            n_jobs=self.n_jobs,
            prefer=self.prefer,
            require=self.require
        )(
            delayed(_parallel_predict)(
                X,
                self.models[i],
                self.model_weights[i]
            )
            for i in range(len(self.model_weights))
        )
        #n_samples, n_models
        preds_proba = np.asarray(preds_proba).sum(axis=0)
        return preds_proba.argmax(axis=1)

    
        


# -*- coding: utf-8 -*-
"""
Created on Wed Jun  8 11:10:32 2022

@author: a694772
"""

import numpy as np
from numba import types, njit
from numba.extending import overload, register_jitable
from numba.core.errors import TypingError

from convst import __USE_NUMBA_CACHE__

@overload(np.all)
def np_all(x, axis=None):

    # ndarray.all with axis arguments for 2D arrays.
    @register_jitable
    def _np_all_axis0(arr):
        out = np.logical_and(arr[0], arr[1])
        for v in iter(arr[2:]):
            for idx, v_2 in enumerate(v):
                out[idx] = np.logical_and(v_2, out[idx])
        return out

    @register_jitable
    def _np_all_flat(x):
        out = x.all()
        return out

    @register_jitable
    def _np_all_axis1(arr):
        out = np.logical_and(arr[:, 0], arr[:, 1])
        for idx, v in enumerate(arr[:, 2:]):
            for v_2 in iter(v):
                out[idx] = np.logical_and(v_2, out[idx])
        return out

    if isinstance(axis, types.Optional):
        axis = axis.type

    if not isinstance(axis, (types.Integer, types.NoneType)):
        raise TypingError("'axis' must be 0, 1, or None")

    if not isinstance(x, types.Array):
        raise TypingError("Only accepts NumPy ndarray")

    if not (1 <= x.ndim <= 2):
        raise TypingError("Only supports 1D or 2D NumPy ndarrays")

    if isinstance(axis, types.NoneType):

        def _np_all_impl(x, axis=None):
            return _np_all_flat(x)

        return _np_all_impl

    elif x.ndim == 1:

        def _np_all_impl(x, axis=None):
            return _np_all_flat(x)

        return _np_all_impl

    elif x.ndim == 2:

        def _np_all_impl(x, axis=None):
            if axis == 0:
                return _np_all_axis0(x)
            else:
                return _np_all_axis1(x)

        return _np_all_impl

    else:

        def _np_all_impl(x, axis=None):
            return _np_all_flat(x)

        return _np_all_impl

@njit(cache=__USE_NUMBA_CACHE__)
def nb_unique(input_data, axis=0):
    """2D np.unique(a, return_index=True, return_counts=True)
    
    Parameters
    ----------
    input_data : 2D numeric array
    axis : int, optional
        axis along which to identify unique slices, by default 0
    Returns
    -------
    2D array
        unique rows (or columns) from the input array
    1D array of ints
        indices of unique rows (or columns) in input array
    1D array of ints
        number of instances of each unique row
    """

    # don't want to sort original data
    if axis == 1:
        data = input_data.T.copy()

    else:
        data = input_data.copy()

    # so we can remember the original indexes of each row
    orig_idx = np.array([i for i in range(data.shape[0])])

    # sort our data AND the original indexes
    for i in range(data.shape[1] - 1, -1, -1):
        sorter = data[:, i].argsort(kind="mergesort")

        # mergesort to keep associations
        data = data[sorter]
        orig_idx = orig_idx[sorter]
    # get original indexes
    idx = [0]

    if data.shape[1] > 1:
        bool_idx = ~np.all((data[:-1] == data[1:]), axis=1)
        additional_uniques = np.nonzero(bool_idx)[0] + 1

    else:
        additional_uniques = np.nonzero(~(data[:-1] == data[1:]))[0] + 1

    idx = np.append(idx, additional_uniques)
    # get counts for each unique row
    counts = np.append(idx[1:], data.shape[0])
    counts = counts - idx
    return data[idx], orig_idx[idx], counts



# -*- coding: utf-8 -*-
"""
Created on Fri Mar 18 16:16:32 2022

@author: a694772
"""
import numpy as np
import pandas as pd
import copy

from sklearn.metrics import f1_score, make_scorer
from sklearn.metrics import accuracy_score
from sklearn.model_selection import cross_validate, GridSearchCV
from sklearn.utils import check_random_state

from timeit import default_timer as timer

from convst.utils.dataset_utils import load_sktime_arff_file_resample_id, load_sktime_dataset_split

#Reuse of the sktime function, modified for numpy array inputs rather than dataframes
# and to handle variable length series
def _sktime_resample(X_train, y_train, X_test, y_test, random_state):
    """Stratified resample data without replacement using a random state.

    Reproducable resampling. Combines train and test, resamples to get the same class
    distribution, then returns new train and test.

    Parameters
    ----------
    X_train : np.array
        train data.
    y_train : np.array
        train data class labels.
    X_test : np.array
        test data.
    y_test : np.array
        test data class labes as np array.
    random_state : int
        seed to enable reproducable resamples
    Returns
    -------
    new train and test attributes and class labels.
    """
    all_labels = np.concatenate((y_train, y_test), axis=None)
    if isinstance(X_train, np.ndarray):
        all_data = np.concatenate([X_train, X_test],axis=0)
    else:
        all_data = X_train + X_test
    random_state = check_random_state(random_state)
    # count class occurrences
    unique_train, counts_train = np.unique(y_train, return_counts=True)
    unique_test, counts_test = np.unique(y_test, return_counts=True)
    assert list(unique_train) == list(
        unique_test
    )  # haven't built functionality to deal with classes that exist in
    # test but not in train
    # prepare outputs
    if isinstance(X_train, np.ndarray):
        X_train = np.empty((0,X_train.shape[1],X_train.shape[2]))
        X_test = np.empty((0,X_test.shape[1],X_test.shape[2]))
    else:
        X_train = []
        X_test = []
    y_train = np.array([])
    y_test = np.array([])
    # for each class
    for label_index in range(0, len(unique_train)):
        # derive how many instances of this class from the counts
        num_instances = counts_train[label_index]
        # get the indices of all instances with this class label
        label = unique_train[label_index]
        indices = np.where(all_labels == label)[0]
        # shuffle them
        random_state.shuffle(indices)
        # take the first lot of instances for train, remainder for test
        train_indices = indices[0:num_instances]
        test_indices = indices[num_instances:]
        del indices  # just to make sure it's not used!
        # extract data from corresponding indices
        # concat onto current data from previous loop iterations
        if isinstance(X_train, np.ndarray):
            train_instances = all_data[train_indices, :]
            test_instances = all_data[test_indices, :]
            X_train = np.concatenate([X_train, train_instances],axis=0)
            X_test = np.concatenate([X_test, test_instances],axis=0)
        else:
            for idx in train_indices:
                X_train.append(all_data[idx])
            for idx in test_indices:
                X_test.append(all_data[idx])
        train_labels = all_labels[train_indices]
        test_labels = all_labels[test_indices]
        y_train = np.concatenate([y_train, train_labels], axis=None)
        y_test = np.concatenate([y_test, test_labels], axis=None)
    # get the counts of the new train and test resample
    unique_train_new, counts_train_new = np.unique(y_train, return_counts=True)
    unique_test_new, counts_test_new = np.unique(y_test, return_counts=True)
    # make sure they match the original distribution of data
    assert list(counts_train_new) == list(counts_train)
    assert list(counts_test_new) == list(counts_test)
    return X_train, y_train, X_test, y_test


class cross_validate_UCR_UEA:
    def __init__(self, n_split, dataset_name, scorers={"accuracy":accuracy_score}):
        self.n_split = n_split
        self.dataset_name = dataset_name
        self.scorers = scorers
        
    def score(self, pipeline):
        X_train_0, X_test_0, y_train_0, y_test_0, _ = load_sktime_dataset_split(
            self.dataset_name
        )
        _score = pd.DataFrame()
        for i in range(self.n_split):
            
            if i == 0:
                if isinstance(X_train_0, np.ndarray):
                    X_train = np.copy(X_train_0)
                    X_test = np.copy(X_test_0)
                    
                else:
                    X_train = copy.copy(X_train_0)
                    X_test = copy.copy(X_test_0)
                y_train = np.copy(y_train_0)
                y_test = np.copy(y_test_0)
            else:
                X_train, y_train, X_test, y_test = _sktime_resample(
                    X_train_0, y_train_0, X_test_0, y_test_0, i
                )
            t0 = timer()
            pipeline = pipeline.fit(X_train, y_train)
            y_pred = pipeline.predict(X_test)
            t1 = timer()
            for scorer_name, scorer in self.scorers.items():
                _score.loc[i, scorer_name] = scorer(y_test, y_pred)
            _score.loc[i, 'time'] = t1 - t0
        return _score
        
    
class _sklearn_sktime_cv:
    def __init__(self, n_splits, dataset_name):
        self.n_splits=n_splits
        self.dataset_name=dataset_name
    
    def split(self, X, y=None, groups=None):
        X_train_0, X_test_0, y_train_0, y_test_0, _ = load_sktime_dataset_split(
            self.dataset_name
        )
        for i in range(self.n_splits):
            if i == 0:
                X_train = np.copy(X_train_0)
                X_test = np.copy(X_test_0)
                y_train = np.copy(y_train_0)
                y_test = np.copy(y_test_0)
            else:
                X_train, y_train, X_test, y_test = _sktime_resample(
                    X_train_0, y_train_0, X_test_0, y_test_0, i
                )
            idx_Train = [np.where((X == X_train[j]).all(axis=2))[0][0] for j in range(X_train.shape[0])]
            idx_Test = [np.where((X == X_test[j]).all(axis=2))[0][0] for j in range(X_test.shape[0])]
            yield idx_Train, idx_Test
    
    def get_n_splits(self, X=None, y=None, groups=None):
        return self.n_splits
    
class grid_search_UCR_UEA:
    def __init__(self, n_split, dataset_name, n_jobs, scorers=make_scorer(accuracy_score)):
        self.n_split = n_split
        self.dataset_name = dataset_name
        self.scorers = scorers
        self.n_jobs=n_jobs
        
    def score(self, pipeline, params):
        cv = GridSearchCV(
            pipeline(), params, scoring=self.scorers, n_jobs=self.n_jobs,
            cv=_sklearn_sktime_cv(self.n_split, self.dataset_name), pre_dispatch='n_jobs',
            verbose=3
        )
        X_train_0, X_test_0, y_train_0, y_test_0, _ = load_sktime_dataset_split(
            self.dataset_name
        )  
        X = np.concatenate((X_train_0, X_test_0),axis=0)
        y = np.concatenate((y_train_0, y_test_0),axis=0)
        cv.fit(X, y)
        return pd.DataFrame(cv.cv_results_)
    
        


class ARFF_stratified_resample:
    """
    Class used as a splitter for sklearn cross validation tools. 
    It will take previsouly resampled arff files at a location and
    return a resample based on the identifier of the current cross
    validation step. 
    
    It is used to reproduce the exact same splits made in the original UCR/UEA
    archive. The arff files can be produced using the tsml java implementation.
    
    Parameters
    ----------
    n_splits : int
        Number of cross validation step planed.
    path : string
        Path to the arff files.
    
    """
    def __init__(self, n_splits, path, normalize=False):
        self.n_splits=n_splits
        self.path=path
        self.normalize=normalize
    
    def split(self, X, y=None, groups=None):
        """
        

        Parameters
        ----------
        X : array, shape=(n_samples, n_features, n_timestamps)
            Time series data to split
        y : ignored
            
        groups : ignored
            

        Yields
        ------
        idx_Train : array, shape(n_samples_train)
            Index of the training data in the original dataset X.
        idx_Test : array, shape(n_samples_test)
            Index of the testing data in the original dataset X.

        """
        for i in range(self.n_splits):
            X_train, X_test, y_train, y_test, _ = load_sktime_arff_file_resample_id(self.path, i, normalize=self.normalize)
            idx_Train = [np.where((X == X_train[j]).all(axis=2))[0][0] for j in range(X_train.shape[0])]
            idx_Test = [np.where((X == X_test[j]).all(axis=2))[0][0] for j in range(X_test.shape[0])]
            yield idx_Train, idx_Test
    
    def get_n_splits(self, X=None, y=None, groups=None):
        """
        Return the number of split made by the splitter. 
        

        Parameters
        ----------
        X : ignored
            
        y : ignored
        
        groups : ignored
            

        Returns
        -------
        n_splits : int
            The n_splits attribute of the object.

        """
        return self.n_splits

# To be used with ARFF_stratified_resample
def run_pipeline(pipeline, X_train, X_test, y_train, y_test, splitter, n_jobs=1):
    """
    Run a sklearn compatible model or pipeline on the specified dataset.

    Parameters
    ----------
    pipeline : object
        A sklearn compatible model or pipeline
    X_train : array, shape=(n_samples, n_feature, n_timestamps)
        Input training data
    X_test : array, shape=(n_samples, n_feature, n_timestamps)
        Input testing data
    y_train : array, shape=(n_samples)
        Class of the input training data.
    y_test : array, shape=(n_samples)
        Class of the input testing data.
    splitter : object
        A sklearn compatible splitter for cross-validation.
    n_jobs : int, optional
        Number of parallel validation round. The default is 1.

    Raises
    ------
    ValueError
        If the number of split of the splitter is an invalid value, raise a
        ValueError exception.

    Returns
    -------
    float
        Mean accuracy over all validation splits
    float
        Std of accuracy over all validation splits
    float
        Mean F1-score over all validation splits
    float
        Std F1-score over all validation splits
    float
        Mean runtime over all validation splits
    float
        Std runtime over all validation splits

    """
    if splitter.n_splits > 1:
        X = np.concatenate([X_train, X_test], axis=0).astype(np.float64)
        y = np.concatenate([y_train, y_test], axis=0).astype(np.int64)
        cv = cross_validate(pipeline, X, y, cv=splitter, n_jobs=n_jobs,
                            scoring={'f1': make_scorer(f1_score, average='macro'),
                                     'acc':make_scorer(accuracy_score)})
        return np.mean(cv['test_acc']), np.std(cv['test_acc']), np.mean(cv['test_f1']), np.std(cv['test_f1']), np.mean(cv['fit_time'] + cv['score_time']), np.std(cv['fit_time'] + cv['score_time'])

    if splitter.n_splits == 1:
        #change datetime to context accurate timing
        t0 = timer()
        pipeline = pipeline.fit(X_train, y_train)
        pred = pipeline.predict(X_test)
        t1 = timer()
        return accuracy_score(y_test, pred), 0, f1_score(y_test, pred, average='macro'), 0, (t1-t0), 0
    
    raise ValueError("Invalid value for n_split in splitter,"
                     " got {} splits".format(splitter.n_splits))



# -*- coding: utf-8 -*-

import numpy as np
from sktime.datasets import (load_UCR_UEA_dataset, 
    load_from_tsfile_to_dataframe, load_from_arff_to_dataframe,
    
)
from sktime.datatypes._panel._convert import (from_multiindex_to_dflist,
    from_nested_to_multi_index
)
from sklearn.preprocessing import LabelEncoder
from numba import njit, prange

from convst import __USE_NUMBA_CACHE__

def _custom_from_nested_to_3d_numpy(X):
    X = from_multiindex_to_dflist(from_nested_to_multi_index(X))
    if all([X[i].shape[0] == X[0].shape[0] for i in range(len(X))]):
        return np.array([X[i].values.T for i in range(len(X))])
    else:
        return [X[i].values.T for i in range(len(X))]

@njit(cache=__USE_NUMBA_CACHE__)
def z_norm_3D(X):
    """
    Z normalise a time series dataset assumed to be of even length. A small value
    is added to the standard deviation for all samples and features to avoid
    0 division.

    Parameters
    ----------
    X : array, shape=(n_samples, n_features, n_timestamps)
        Input numerical array to z-normalise

    Returns
    -------
    X : array, shape=(n_samples, n_features, n_timestamps)
        Z-normalised array

    """
    for i_x in prange(X.shape[0]):
        for i_ft in prange(X.shape[1]):
            X[i_x, i_ft] = (X[i_x, i_ft] - X[i_x, i_ft].mean())/(X[i_x, i_ft].std() + 1e-8)
    return X
    
def z_norm_3D_list(X):
    """
    Z normalise a time series dataset assumed to be of even length. A small value
    is added to the standard deviation for all samples and features to avoid
    0 division.

    Parameters
    ----------
    X : array, shape=(n_samples, n_features, n_timestamps)
        Input numerical array to z-normalise

    Returns
    -------
    X : array, shape=(n_samples, n_features, n_timestamps)
        Z-normalised array

    """
    for i_x in range(len(X)):
        for i_ft in range(len(X[i_x])):
            X[i_x][i_ft] = (X[i_x][i_ft] - X[i_x][i_ft].mean())/(X[i_x][i_ft].std() + 1e-8)
    return X
    

def load_sktime_dataset_split(name, normalize=False):
    """
    Load the original train and test splits of a dataset 
    from the UCR/UEA archive by name using sktime API.

    Parameters
    ----------
    name : string
        Name of the dataset to download.
    normalize : boolean, optional
        If True, time series will be z-normalized. The default is True.


    Returns
    -------
    X_train : array, shape=(n_samples_train, n_features, n_timestamps)
        Training data from the dataset specified by path.
    X_test : array, shape=(n_samples_test, n_features, n_timestamps)
        Testing data from the dataset specified by path.
    y_train : array, shape=(n_samples_train)
        Class of the training data.
    y_test : array, shape=(n_samples_test)
        Class of the testing data.
    le : LabelEncoder
        LabelEncoder object used to uniformize the class labels


    """
    #Load datasets
    X_train, y_train = load_UCR_UEA_dataset(name, return_X_y=True, split='train')
    X_test, y_test = load_UCR_UEA_dataset(name, return_X_y=True, split='test')

    #Convert pandas DataFrames to numpy arrays
    X_train = _custom_from_nested_to_3d_numpy(X_train)
    X_test = _custom_from_nested_to_3d_numpy(X_test)

    #Convert class labels to make sure they are between 0,n_classes
    le = LabelEncoder().fit(y_train)
    y_train = le.transform(y_train)
    y_test = le.transform(y_test)
    min_len = min(
        min([len(X_train[i][0])for i in range(len(X_train))]),
        min([len(X_test[i][0])for i in range(len(X_test))]),
    )
    #Z-Normalize the data
    if normalize and not isinstance(X_train, list):
        X_train = z_norm_3D(X_train)
        X_test = z_norm_3D(X_test)
    if normalize and isinstance(X_train, list):
        X_train = z_norm_3D_list(X_train)
        X_test = z_norm_3D_list(X_test)
    return X_train, X_test, y_train, y_test, min_len


def load_sktime_arff_file(path, normalize=False):
    """
    Load a dataset from .arff files.

    Parameters
    ----------
    path : string
        Path to the folder containing the .ts file. Dataset name
        should be specified at the end of path to find files as
        "dataset_TRAIN.arff" and "dataset_TEST.arff"
    normalize : boolean, optional
        If True, time series will be z-normalized. The default is True.


    Returns
    -------
    X_train : array, shape=(n_samples_train, n_features, n_timestamps)
        Training data from the dataset specified by path.
    X_test : array, shape=(n_samples_test, n_features, n_timestamps)
        Testing data from the dataset specified by path.
    y_train : array, shape=(n_samples_train)
        Class of the training data.
    y_test : array, shape=(n_samples_test)
        Class of the testing data.
    le : LabelEncoder
        LabelEncoder object used to uniformize the class labels

    """
    #Load datasets
    X_train, y_train = load_from_arff_to_dataframe(path+'_TRAIN.arff')
    X_test, y_test = load_from_arff_to_dataframe(path+'_TEST.arff')

    #Convert pandas DataFrames to numpy arrays
    X_train = _custom_from_nested_to_3d_numpy(X_train)
    X_test = _custom_from_nested_to_3d_numpy(X_test)

    #Convert class labels to make sure they are between 0,n_classes
    le = LabelEncoder().fit(y_train)
    y_train = le.transform(y_train)
    y_test = le.transform(y_test)

    #Z-Normalize the data
    if normalize:
        X_train = (X_train - X_train.mean(axis=-1, keepdims=True)) / (
            X_train.std(axis=-1, keepdims=True) + 1e-8)
        X_test = (X_test - X_test.mean(axis=-1, keepdims=True)) / (
            X_test.std(axis=-1, keepdims=True) + 1e-8)

    return X_train, X_test, y_train, y_test, le


def load_sktime_arff_file_resample_id(path, rs_id, normalize=False):
    """
    Load a dataset resample from .arff files and the identifier of the 
    resample.

    Parameters
    ----------
    path : string
        Path to the folder containing the .ts file. Dataset name
        should be specified at the end of path to find files as
        "dataset_{rs_id}_TRAIN.arff" and "dataset_{rs_id}_TEST.arff"
    rs_id : int or str
        Identifier of the resample.
    normalize : boolean, optional
        If True, time series will be z-normalized. The default is True.


    Returns
    -------
    X_train : array, shape=(n_samples_train, n_features, n_timestamps)
        Training data from the dataset specified by path.
    X_test : array, shape=(n_samples_test, n_features, n_timestamps)
        Testing data from the dataset specified by path.
    y_train : array, shape=(n_samples_train)
        Class of the training data.
    y_test : array, shape=(n_samples_test)
        Class of the testing data.
    le : LabelEncoder
        LabelEncoder object used to uniformize the class labels

    """
    #Load datasets
    X_train, y_train = load_from_arff_to_dataframe(path+'_{}_TRAIN.arff'.format(rs_id))
    X_test, y_test = load_from_arff_to_dataframe(path+'_{}_TEST.arff'.format(rs_id))

    #Convert pandas DataFrames to numpy arrays
    X_train = _custom_from_nested_to_3d_numpy(X_train)
    X_test = _custom_from_nested_to_3d_numpy(X_test)

    #Convert class labels to make sure they are between 0,n_classes
    le = LabelEncoder().fit(y_train)
    y_train = le.transform(y_train)
    y_test = le.transform(y_test)

    #Z-Normalize the data
    if normalize:
        X_train = (X_train - X_train.mean(axis=-1, keepdims=True)) / (
            X_train.std(axis=-1, keepdims=True) + 1e-8)
        X_test = (X_test - X_test.mean(axis=-1, keepdims=True)) / (
            X_test.std(axis=-1, keepdims=True) + 1e-8)

    return X_train, X_test, y_train, y_test, le

def load_sktime_ts_file(path, normalize=False):
    """
    Load a dataset from .ts files

    Parameters
    ----------
    path : string
        Path to the folder containing the .ts file. Dataset name
        should be specified at the end of path to find files as
        "dataset_TRAIN.ts" and "dataset_TEST.ts"
    normalize : boolean, optional
        If True, time series will be z-normalized. The default is True.

    Returns
    -------
    X_train : array, shape=(n_samples_train, n_features, n_timestamps)
        Training data from the dataset specified by path.
    X_test : array, shape=(n_samples_test, n_features, n_timestamps)
        Testing data from the dataset specified by path.
    y_train : array, shape=(n_samples_train)
        Class of the training data.
    y_test : array, shape=(n_samples_test)
        Class of the testing data.
    le : LabelEncoder
        LabelEncoder object used to uniformize the class labels

    """
    
    #Load datasets
    X_train, y_train = load_from_tsfile_to_dataframe(path+'_TRAIN.ts')
    X_test, y_test = load_from_tsfile_to_dataframe(path+'_TEST.ts')

    #Convert pandas DataFrames to numpy arrays
    X_train = _custom_from_nested_to_3d_numpy(X_train)
    X_test = _custom_from_nested_to_3d_numpy(X_test)

    #Convert class labels to make sure they are between 0,n_classes
    le = LabelEncoder().fit(y_train)
    y_train = le.transform(y_train)
    y_test = le.transform(y_test)

    #Z-Normalize the data
    if normalize:
        X_train = (X_train - X_train.mean(axis=-1, keepdims=True)) / (
            X_train.std(axis=-1, keepdims=True) + 1e-8)
        X_test = (X_test - X_test.mean(axis=-1, keepdims=True)) / (
            X_test.std(axis=-1, keepdims=True) + 1e-8)

    return X_train, X_test, y_train, y_test, le

def load_sktime_dataset(name, normalize=False):
    """
    Load a dataset from the UCR/UEA archive by name using sktime API

    Parameters
    ----------
    name : string
        Name of the dataset to download.
    normalize : boolean, optional
        If True, time series will be z-normalized. The default is True.

    Returns
    -------
    X : array, shape=(n_samples, n_features, n_timestamps)
        Time series data from the dataset specified by name.
    y : array, shape=(n_samples)
        Class of the time series
    le : LabelEncoder
        LabelEncoder object used to uniformize the class labels
        

    """
    #Load datasets
    X_train, X_test, y_train, y_test, le = load_sktime_dataset_split(
        name, normalize=normalize
    )
    
    return np.concatenate((X_train, X_test),axis=0), np.concatenate((y_train, y_test),axis=0), le

def return_all_dataset_names():
    return np.concatenate((
        return_all_univariate_dataset_names(),
        return_all_multivariate_dataset_names(),
        return_all_variable_univariate_dataset_names(),
        return_all_variable_multivariate_dataset_names()
    ))

def return_all_multivariate_dataset_names():
    
    return np.asarray([
        "ArticularyWordRecognition",
        "AtrialFibrillation",
        "BasicMotions",
        "Cricket",
        "DuckDuckGeese",
        "EigenWorms",
        "Epilepsy",
        "EthanolConcentration",
        "ERing",
        "FaceDetection",
        "FingerMovements",
        "HandMovementDirection",
        "Handwriting",
        "Heartbeat",
        "Libras",
        "LSST",
        "MotorImagery",
        "NATOPS",
        "PenDigits",
        "PEMS-SF",
        "PhonemeSpectra",
        "RacketSports",
        "SelfRegulationSCP1",
        "SelfRegulationSCP2",
        "StandWalkJump",
        "UWaveGestureLibrary"
    ])

def return_all_variable_multivariate_dataset_names():
    return np.asarray([
        "AsphaltObstaclesCoordinates",
        "AsphaltPavementTypeCoordinates",
        "AsphaltRegularityCoordinates",
        "CharacterTrajectories",
        "InsectWingbeat",
        "JapaneseVowels",
        "SpokenArabicDigits"
    ])


def return_all_variable_univariate_dataset_names():
    return np.asarray([
        "AllGestureWiimoteX",
        "AllGestureWiimoteY",
        "AllGestureWiimoteZ",
        "GestureMidAirD1",
        "GestureMidAirD2",
        "GestureMidAirD3",
        "GesturePebbleZ1",
        "GesturePebbleZ2",
        "PickupGestureWiimoteZ",
        "PLAID",
        "ShakeGestureWiimoteZ"
    ])


def return_all_univariate_dataset_names():
    """
    Return the names of the 112 univariate datasets of the UCR archive.

    Returns
    -------
    array, shape=(112)
        Names of the univariate UCR datasets.

    """
    return np.asarray([
            "ACSF1",
            "Adiac",
            "ArrowHead",
            "Beef",
            "BeetleFly",
            "BirdChicken",
            "BME",
            "Car",
            "CBF",
            "Chinatown",
            "ChlorineConcentration",
            "Coffee",
            "Computers",
            "CricketX",
            "CricketY",
            "CricketZ",
            "DiatomSizeReduction",
            "DistalPhalanxOutlineAgeGroup",
            "DistalPhalanxOutlineCorrect",
            "DistalPhalanxTW",
            "Earthquakes",
            "ECG200",
            "ECG5000",
            "ECGFiveDays",
            "ElectricDevices",
            "EOGHorizontalSignal",
            "EOGVerticalSignal",
            "FaceAll",
            "FaceFour",
            "FacesUCR",
            "FiftyWords",
            "Fish",
            "FordA",
            "FordB",
            "FreezerRegularTrain",
            "FreezerSmallTrain",
            "GunPoint",
            "GunPointAgeSpan",
            "GunPointMaleVersusFemale",
            "GunPointOldVersusYoung",
            "Ham",
            "Haptics",
            "Herring",
            "HouseTwenty",
            "InlineSkate",
            "InsectEPGRegularTrain",
            "InsectEPGSmallTrain",
            "InsectWingbeatSound",
            "ItalyPowerDemand",
            "LargeKitchenAppliances",
            "Lightning2",
            "Lightning7",
            "Mallat",
            "Meat",
            "MedicalImages",
            "MiddlePhalanxOutlineAgeGroup",
            "MiddlePhalanxOutlineCorrect",
            "MiddlePhalanxTW",
            "MixedShapesRegularTrain",
            "MixedShapesSmallTrain",
            "MoteStrain",
            "NonInvasiveFetalECGThorax1",
            "NonInvasiveFetalECGThorax2",
            "OliveOil",
            "OSULeaf",
            "PhalangesOutlinesCorrect",
            "Plane",
            "PowerCons",
            "ProximalPhalanxOutlineAgeGroup",
            "ProximalPhalanxOutlineCorrect",
            "ProximalPhalanxTW",
            "RefrigerationDevices",
            "ScreenType",
            "SemgHandGenderCh2",
            "SemgHandMovementCh2",
            "SemgHandSubjectCh2",
            "ShapeletSim",
            "ShapesAll",
            "SmallKitchenAppliances",
            "SmoothSubspace",
            "SonyAIBORobotSurface1",
            "SonyAIBORobotSurface2",
            "Strawberry",
            "SwedishLeaf",
            "Symbols",
            "SyntheticControl",
            "ToeSegmentation1",
            "ToeSegmentation2",
            "Trace",
            "TwoLeadECG",
            "TwoPatterns",
            "UMD",
            "UWaveGestureLibraryAll",
            "UWaveGestureLibraryX",
            "UWaveGestureLibraryY",
            "UWaveGestureLibraryZ",
            "Wafer",
            "Wine",
            "WordSynonyms",
            "Worms",
            "WormsTwoClass",
            "Yoga",
            "Phoneme",
            "PigAirwayPressure",
            "PigArtPressure",
            "PigCVP",
            "Crop",
            "StarLightCurves",
            "Rock",
            "HandOutlines",
            "CinCECGTorso",
            "EthanolLevel"
            ])



# -*- coding: utf-8 -*-
"""
Created on Tue Sep 27 20:26:34 2022

@author: antoi
"""
import pytest

import numpy as np

from convst.classifiers import R_DST_Ridge
from convst.transformers import R_DST
from convst.utils.dataset_utils import load_sktime_dataset_split
from convst.utils.experiments_utils import cross_validate_UCR_UEA
from convst.transformers._commons import is_prime

import logging

LOGGER = logging.getLogger(__name__)

@pytest.mark.parametrize("name, expected", [
    ('GunPoint','univariate'),
    ('BasicMotions','multivariate'),
    ('PLAID','univariate_variable'),
    ('AsphaltObstaclesCoordinates','multivariate_variable')
])
def test_auto_type(name, expected):
    X_train, X_test, y_train, y_test, min_len = load_sktime_dataset_split(
        name=name
    )
    rdst = R_DST(n_shapelets=2, min_len=min_len).fit(X_train, y_train)
    assert rdst.transform_type == expected


@pytest.mark.parametrize("name, lengths", [
    ('GunPoint',[11]),
    ('GunPoint',[0.05]),
    ('GunPoint',[11,15,19]),
    ('GunPoint',[0.05,0.08,0.1]),
    ('BasicMotions',[11]),
    ('BasicMotions',[0.05]),
    ('BasicMotions',[11,15,19]),
    ('BasicMotions',[0.05,0.08,0.1]),
    ('PLAID',[11]),
    ('PLAID',[0.05]),
    ('PLAID',[11,15,19]),
    ('PLAID',[0.05,0.08,0.1]),
    ('AsphaltObstaclesCoordinates',[11]),
    ('AsphaltObstaclesCoordinates',[0.05]),
    ('AsphaltObstaclesCoordinates',[11,15,19]),
    ('AsphaltObstaclesCoordinates',[0.05,0.08,0.1])
])
def test_mutliple_lengths(name, lengths):
    X_train, X_test, y_train, y_test, min_len = load_sktime_dataset_split(
        name=name
    )
    try:
        R_DST(n_shapelets=1000, shapelet_lengths=lengths, min_len=min_len).fit(X_train, y_train)
    except Exception as e:
        LOGGER.info('A data format test failed on {} due to the following exception : {}'.format(
             name, e  
        ))
        assert False 
    assert True

@pytest.mark.parametrize("name", [
    ('FordA'),
    ('FordB')
])
def test_prime_dilations(name):
    X_train, X_test, y_train, y_test, min_len = load_sktime_dataset_split(
        name=name
    )
    try:
        rdst = R_DST_Ridge(min_len=min_len, prime_dilations=True).fit(X_train, y_train)
        rdst.score(X_test, y_test)
    except Exception as e:
        LOGGER.info('An exception as occured during prime dilation tests: {}'.format(
            e  
        ))
        assert False 
    assert all([is_prime(i) for i in rdst.transformer.shapelets_[2]])


# TODO : this may fail due to unlucky generation of shapelets, if a length 
# is not selected randomly, the expected array will be bigger than actual
@pytest.mark.parametrize("name, bounds, reduction, expected", [
    ('GunPoint', [6, 12], 0., [6, 7, 8, 9, 10, 11, 12]),
    ('GunPoint', [6, 12], 0.5, [6, 8, 10, 12]),
    ('GunPoint', [0.1, 0.15], 0., [15, 16, 17, 18, 19, 20, 21, 22]),
    ('GunPoint', [0.1, 0.15], 0.5, [15, 17, 19, 21])
])
def test_length_bounds(name, bounds, reduction, expected):
    X_train, X_test, y_train, y_test, min_len = load_sktime_dataset_split(
        name=name
    )
    try:
        rdst = R_DST_Ridge(
            min_len=min_len, shapelet_lengths_bounds=bounds, 
            lengths_bounds_reduction=reduction
        ).fit(X_train, y_train)
        rdst.score(X_test, y_test)
    except Exception as e:
        LOGGER.info('An exception as occured during length bounds tests: {}'.format(
            e  
        ))
        assert False 
    assert list(np.unique(rdst.transformer.shapelets_[1])) == expected

# Lower than actual best accuracy to account for possible deviation due to random sampling
@pytest.mark.parametrize("name, expected", [
    ('GunPoint',0.98),
    ('Wine',0.94),
    ('BasicMotions',0.94),
    ('PLAID',0.89),
    ('AsphaltObstaclesCoordinates',0.78)
])
def test_performance(name, expected):
    X_train, X_test, y_train, y_test, min_len = load_sktime_dataset_split(
        name=name
    )
    rdst = R_DST_Ridge(n_shapelets=1,min_len=min_len).fit(X_train, y_train)
    rdst.score(X_test, y_test)
    assert rdst.transformer.shapelets_[1].shape[0] == 1
    acc = cross_validate_UCR_UEA(5,name).score(R_DST_Ridge(n_jobs=-1,min_len=min_len))
    acc = acc['accuracy'].mean()
    LOGGER.info('{} Dataset -> Accuracy {}, Expected >= {}'.format(
         name, acc, expected   
    ))
    assert acc >= expected


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Sep  6 16:35:58 2022

@author: lifo
"""
import numpy as np


from convst.utils.checks_utils import check_array_3D

from numba import njit, prange

from pyts.approximation import DiscreteFourierTransform, SymbolicAggregateApproximation

from scipy.signal import periodogram
from scipy.fft import fht, fhtoffset

from sklearn.preprocessing import StandardScaler, MinMaxScaler
from sklearn.base import BaseEstimator, TransformerMixin


from convst import (
    __USE_NUMBA_CACHE__, __USE_NUMBA_PARALLEL__
)


class c_StandardScaler(StandardScaler):
    def fit(self, X, y=None):
        self.usefull_atts = np.where(np.std(X, axis=0) != 0)[0]
        return super().fit(X[:, self.usefull_atts], y=y)
    
    def transform(self, X):
        return super().transform(X[:, self.usefull_atts])
    
class c_MinMaxScaler(MinMaxScaler):
    def fit(self, X, y=None):
        self.usefull_atts = np.where(np.std(X, axis=0) != 0)[0]
        return super().fit(X[:, self.usefull_atts], y=y)
    
    def transform(self, X):
        return super().transform(X[:, self.usefull_atts])

@njit(cache=__USE_NUMBA_CACHE__)
def z_norm_one_sample(x):
    n_features, n_timestamps = x.shape
    x_new = np.empty((n_features, n_timestamps))
    for i in prange(n_features):
        x_new[i] = (x[i] - x[i].mean()) / (x[i].std() + 1e-8)
    return x_new

@njit(cache=__USE_NUMBA_CACHE__, parallel=__USE_NUMBA_PARALLEL__)
def z_norm_all_samples(X):
    n_samples, n_features, n_timestamps = X.shape
    X_new = np.empty((n_samples, n_features, n_timestamps))
    for i in prange(n_samples):
        X_new[i] = z_norm_one_sample(X[i])
    return X_new

class Z_normalizer(BaseEstimator, TransformerMixin):
    def __init__(self):
        self.is_univariate=False
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        X = check_array_3D(X, is_univariate=self.is_univariate)
        return z_norm_all_samples(X)
    
class Raw(BaseEstimator, TransformerMixin):
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        return X

class Derivate(BaseEstimator, TransformerMixin):
    def __init__(self, order=1, random=False):
        self.order = order
        self.random = random
        
    def fit(self, X, y=None):
        if self.random:
            self._random_init()
        return self
    
    def transform(self, X):
        for i in range(self.order):
            X = np.diff(X)
        return X

    def _random_init(self):
        self.set_params(**{"order":np.random.choice(np.arange(1,5))})

class Periodigram(BaseEstimator, TransformerMixin):
    def __init__(self, window_type="boxcar", random=False):
        self.window_type = window_type
        self.random = random
        self.is_univariate=False
        
    def fit(self, X, y=None):
        if self.random:
            self._random_init()
        return self
    
    def transform(self, X):
        n_ts = periodogram(X[0,0,:], detrend=False, window=self.window_type)[1].shape[0]
        X_new = np.empty((X.shape[0], X.shape[1], n_ts))
        for i in range(X.shape[0]):
            for j in range(X.shape[1]):
                X_new[i,j] = periodogram(X[i,j], detrend=False, window=self.window_type)[1]
        return X_new
    
    def _random_init(self):
        self.set_params(**{"window_type":np.random.choice(self._get_windows())})
    
    def _get_windows(self):
        return np.asarray(
            ["boxcar","triang","blackman","hamming","hann",
             "bartlett","flattop","parzen","bohman",
             "blackmanharris","nuttall","barthann",
             "cosine","exponential","tukey","taylor"]
        )

#TODO : adapt to multivariate/uneven length context
class Sax(BaseEstimator, TransformerMixin):
    def __init__(self, n_bins=10, strategy="uniform", random=False):
        self.random = random
        self.n_bins = n_bins
        self.strategy = strategy
        
    
    def fit(self, X, y=None):
        if self.random:
            self._random_init(X.shape[1])
        self.transformer = SymbolicAggregateApproximation(
            n_bins=self.n_bins, strategy=self.strategy, alphabet='ordinal'
        )
        self.transformer.fit(X[:,0,:])
        return self
    
    def transform(self, X):
        X = self.transformer.transform(X[:,0,:])
        return X[:, np.newaxis, :]
    
    def _random_init(self, n_timestamps):
        self.set_params(**{"n_bins":np.random.choice(np.arange(2,min(n_timestamps,26)))})

#TODO : adapt to multivariate/uneven length context
class FourrierCoefs(BaseEstimator, TransformerMixin):
    def __init__(
        self, n_coefs=None, drop_sum=False, anova=False, norm_mean=False,
        norm_std=False
    ):
        self.n_coefs = n_coefs
        self.drop_sum = drop_sum
        self.anova = anova
        self.norm_mean = norm_mean
        self.norm_std = norm_std
    
    def fit(self, X, y=None):
        self.transformer = DiscreteFourierTransform(
            n_coefs=self.n_coefs, drop_sum=self.drop_sum, anova=self.anova,
            norm_std=self.norm_std, norm_mean=self.norm_mean,
        )
        self.transformer.fit(X[:,0,:], y=y)
        return self
    
    def transform(self, X):
        X = self.transformer.transform(X[:,0,:])
        return X[:, np.newaxis, :]


class FastHankelTransform(BaseEstimator, TransformerMixin):
    def __init__(
        self, dln=0.01, mu=1, offset=0.0, bias=0.0, use_optimal_offset=True
    ):
        self.dln = dln
        self.mu = mu
        self.offset = offset
        self.bias = bias
        self.use_optimal_offset = use_optimal_offset
    
    def fit(self, X, y=None):
        return self
    
    def transform(self, X):
        X_new = np.zeros(X.shape)
        for i in range(X.shape[0]):
            if self.use_optimal_offset:
                X_new[i] = fht(
                    X[i], self.dln, self.mu,
                    offset=fhtoffset(self.dln, self.dln, self.bias),
                    bias=self.bias
                )
            else:
                X_new[i] = fht(
                    X[i], self.dln, self.mu, offset=self.offset,
                    bias=self.bias
                )
        return X_new


from convst.utils.checks_utils import check_array_3D, check_array_2D, check_array_1D
import numpy as np
import pytest


def init_numpy(dims):
    return np.random.random_sample(dims)
    
##########################################
#                                        #
#             Test 3D checks             #
#                                        #
##########################################

@pytest.mark.parametrize("dims", [
    ((30, 15, 12)),
    ((2, 1, 18)),
])
def test_check_3D_numpy_no_change(dims):
    X = init_numpy(dims)
    X2 = check_array_3D(X)
    assert np.array_equal(X,X2)


@pytest.mark.parametrize("dims", [
    ((30, 15)),
    ((30)),
])
def test_check_3D_numpy_wrong_dimensions(dims):
    X = init_numpy(dims)
    with pytest.raises(ValueError):
        check_array_3D(X)
    
@pytest.mark.parametrize("dims", [
    ((30, 0, 15)),
    ((30, 15, 0)),
    ((0, 15, 30)),
])
def test_check_3D_numpy_empty_dimensions(dims):
    X = init_numpy(dims)
    with pytest.raises(ValueError):
        check_array_3D(X)

##########################################
#                                        #
#             Test 2D checks             #
#                                        #
##########################################  

@pytest.mark.parametrize("dims", [
    ((30, 15)),
    ((2, 1)),
])
def test_check_2D_numpy_no_change(dims):
    X = init_numpy(dims)
    X2 = check_array_2D(X)
    assert np.array_equal(X,X2)


@pytest.mark.parametrize("dims", [
    ((30, 15, 20)),
    ((30)),
])
def test_check_2D_numpy_wrong_dimensions(dims):
    X = init_numpy(dims)
    with pytest.raises(ValueError):
        check_array_2D(X)
    
@pytest.mark.parametrize("dims", [
    ((30, 0)),
    ((0, 15)),
])
def test_check_2D_numpy_empty_dimensions(dims):
    X = init_numpy(dims)
    with pytest.raises(ValueError):
        check_array_2D(X)
        
##########################################
#                                        #
#             Test 1D checks             #
#                                        #
##########################################        

@pytest.mark.parametrize("dims", [
    ((30)),
    ((2)),
])
def test_check_1D_numpy_no_change(dims):
    X = init_numpy(dims)
    X2 = check_array_1D(X)
    assert np.array_equal(X,X2)


@pytest.mark.parametrize("dims", [
    ((30, 15)),
    ((30, 15, 30)),
])
def test_check_1D_numpy_wrong_dimensions(dims):
    X = init_numpy(dims)
    with pytest.raises(ValueError):
        check_array_1D(X)


@pytest.mark.parametrize("dims", [
    ((0)),
])
def test_check_1D_numpy_empty_dimensions(dims):
    X = init_numpy(dims)
    with pytest.raises(ValueError):
        check_array_1D(X)
        

##########################################
#                                        #
#            Test type checks            #
#                                        #
##########################################  
    
@pytest.mark.parametrize("X", [
    ("3"),
    (1.0),
    (1),
    ((1,1,1)),
])
def test_type_check(X):
    with pytest.raises(ValueError):
        check_array_1D(X)
        
